/*
 * synaptics.c - support for the synaptics serial and ps2 touchpads
 *
 * Copyright 1999   hdavies@ameritech.net (Henry Davies)
 *                  Geert Van der Plas provided the code to support
 *                        older Synaptics PS/2 touchpads.
 *
 *   This program is free software; you can redistribute it and/or modify
 *   it under the terms of the GNU General Public License as published by
 *   the Free Software Foundation; either version 2 of the License, or
 *   (at your option) any later version.
 *
 *   This program is distributed in the hope that it will be useful,
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *   GNU General Public License for more details.
 *
 *   You should have received a copy of the GNU General Public License
 *   along with this program; if not, write to the Free Software
 *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.
 *
 ********/

/*
** Design/Goals
**     I want to use the Synaptics Serial touchpad as a cursor device under
**     Linux (gpm).  With this device I want to support operations similar
**     to those supported by the Synaptics WinDOS driver, and some others 
**     of my own devising.
**
** Features:
**     Corner Clicks
**         This recognizes taps on the touchpad in the corner(s) and
**         translates them into specific actions.  Initially I am looking
**         at actions on the order of alternate button clicks.  Other 
**         alternatives include drags and whatnot.
**     Edge Extensions
**         This recognizes that the finger has moved from the center region
**         of the touchpad and dragged to the edge area.  At which point
**         I want to be able to extend the motion by automatically moving 
**         in the direction of the edge.
**     Toss n Catch
**         This recognizes a quick motion of the finger on the touchpad and
**         uses that to define a velocity vector for the cursor.  A tap
**         on the touchpad at a later time catches (stops) the cursor.
**     Tap n Drag
**         A quick tap of the touchpad followed by finger motion on the 
**         touchpad initiates what would be a drag with a normal mouse
**         type device.
**     Pressure Sensitive Velocity
**         The Synaptics touchpad indicates the touch pressure of the finger
**         (really an interface area) this is used to accelerate the cursor
**         motion.  This can be used in the normal motion, Tap n Drag, or
**         Edge Extension modes.  In normal motion and Tap n Drag this may 
**         be awkward due to increased friction caused by the pressure.
**
** Parameters:
**     search below for configuration constants
*/


/* Pebl - pebl@diku.dk 20/12/2001 08:06.
**
** I want to have more features :). Every touch pad (ps2 and serial) is now be
** supported according to STIG and every feature/capability is read.  However
** there seems to be some very old serial touchpads that have an advanced
** mode not mentioned in STIG.
**
**
** Further I have added the following
**
**  Emulation of scrolling
**        The window drivers allow using the edge to emulate a wheel
**        mouse. When putting the finger at the right edge, a movement up or
**        down translates to wheel movement. Taking the finger to the
**        top/button edge keeps the wheel turning.  Lift the finger again to
**        operate normally again. Another option is to do a toss scrolling.
**
**  Stick attached to a synaptic touchpad (aka styk)
**        Some touchpads have a stick "attached", so they share the same port.
**        In absolute mode the stick protocol is a simple ps2 protocol, except
**        it is sent in an absolute packet. (In relative mode both uses ps2)
**        Some sticks send packets when pressed (different from moved), which
**        will be reported as left button.
**
**  4 Way button attached to a touchpad
**        Likewise some touchpads have a round button giving the choices of
**        four way to press the button. The packet is nearly identical to the
**        stick, turn off the stick if you have this button. This is no longer
**        needed.  The button can act as a mouse or as buttons which can be
**        changed on the fly. The buttons can be configured just as the
**        touchpads corners.
**
**  Four buttons to work
**        Some touchpads have 4 buttons. Only 3 was read and the last was set to 
**        up or down "button". 
**
**  Multiple fingers
**        I have added an option to detect such by looking at pressure levels,
**        which is not that great. If the pad have the capability to detect
**        it, this is used instead, but this far from optimal. It does detect
**        two horizontal parted finger if in same vertical position, otherwise
**        the detection is bad. This is/was a problem as my wrist lays on the
**        laptop which gives a 45 deg to horizontal for my fingers. Now I try
**        to add further detections, which works for me in 95% of the time.
**
**  Multi finger tap
**        Using 1,2 or 3 fingers to make a tap translates into left, right or
**        middle button pressed. I originally thought about doing what is
**        called a "HOP" where after a quick finger shift the distance of the
**        hop decides which button is pressed. I dropped it partly because
**        synaptics did in version 3.2 and because it was harder than I
**        thought.
**
**  Multi finger less sensitive
**        Adding or removing a finger or just accidently touch the pad with
**        the palm while using the pad causes the pad to report the average
**        position between the touched places.  This of cause gives annoying
**        erratic movements. When adding or removing fingers are detected
**        the mouse is stop for some small time to avoid this. This is not
**        perfect as the finger detection is not perfect, but it helps a lot!
**
**  Palm detection
**        Some touchpads tries to detect whether the reported data is from a
**        palm or similar.  When reading this signal the movement is
**        stopped. Except the detection does not work very well :(
**
**
**  Enabling/disabling touchpad
**        A new corner action is to disable the touchpad. It is enabled again
**        by tapping the same corner. This is useful if I know I am going
**        to write a lot. The stick still works.
**
**  Debugging corner action  
**        A corner action now allows toggling the debug information. 
**
**
** The structure has changed to handle both serial and ps2 touchpads. Many
** variable names have changed.                  
**
** The documentation referred to is from
**  "Synaptics TouchPad Interfacing Guide" revision 2.5 
** or just STIG
**                                                                                 
**/


/*
** TODO
**   - handle other versions of synaptics touchpads (mine is 3.4 firmware)
**     (this should be close)
**   - test this with more Synaptics touch pads (laptops and such)
**   - provide a configuration interface to adjust parameters and
**     enabled features
**   - determine appropriate ranges for adjusting parameters
**
** TODO (pebl)
**   - Two pads can not be used at the same time (internal/external)
**   - Move a lot of the comments to a readme file.
**   - Allow normal taps in unused corners. 
**   - Better detection of spontaneous reseting touchpad.
**   - Implement resend command for serial connection.
**   - Only start scrolling if mostly vertical movement.
**   - Disable/enable command for serial touchpad.
**   - Clean up the mess about extended packets with pressure 0.
**   - source splitting of normal touchpad functions and synaptic specific and
**     add other touchpads.
*/



/**
** Notation:
** A gesture means a motion or action that is not a regular mouse movement.
** 
** Wmode is an absolute mode where no gesture is signaled by the touchpad, only
** supported on newer versions.
**
** variables *_enabled are use for enable behaviors that may not supported in
** the touchpad.
**
** Many variables have two counterparts (and some that dont should have): a
** *_time and *_packet, where the first is in msec and the latter is the same
** time converted to the number of packets, when using 80 packets per second.
**
** Variables:
** last_* is for last reviewed packet, not last reported/calculated movement/event.
** was_*  is for last event, not necessary last packet.
**
**
** Data process line:
**
** syn_process_ps2_data   ---- syn_translate_serial_data    --                         -----------------------------------------------
**                                (handles also wmode)       |                         |                            |
**                                                           |                         |                            V
**                                                           |- syn_preprocess_report --- syn_process_wmode_report --- syn_process_report
**                                                           |                          
** syn_process_serial_data --- syn_translate_ps2_wmode_data --                                                       
**                          |                                |                          
**                          |                                |                          
**                          -- syn_translate_ps2_data     ----                          
**
**
*/


#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <sys/time.h>
#include <sys/ioctl.h>

#include "headers/gpm.h"
#include "headers/gpmInt.h"       /* which_mouse */
#include "headers/synaptics.h"
#include "headers/message.h"

#define DEBUG                     1
#undef DEBUG

#ifdef DEBUG
#  define DEBUG_SYNAPTIC          
#  define DEBUG_PARAMS            1
#  define DEBUG_FLUSH             1
#  define DEBUG_RESET             1
#  define DEBUG_SENT_DATA         1
#  define DEBUG_RECEIVED_DATA     1
#  define DEBUG_TOSS              1
#  define DEBUG_PALM              1
#  define DEBUG_STICK             1
#  define DEBUG_MULTI_FINGER      1
#  define DEBUG_REPORTS           1    /* this is VERY verbose */
#endif


/* The next is UGLY, but I only want debug info from this file */
/* it's really ugly and you shouldn't use this. Instead use
 * gpm -D | grep filename.c.
 * ifdef DEBUG should not be used anymore! */
/* pebl: But this require that I boot my machine with gpm -D which I dont. The
   intention is that when I am doing my normal doing, and the touchpad suddenly
   behaves strange, I can start debuging without restarting gpm, using a
   corneraction.
*/

/* BAD CODE BEGIN 
#ifdef DEBUG_SYNAPTIC
#  include <stdarg.h>
   static int debug_syn_to_stderr = 1;

   static void gpm_report_static(int line, char *file, int stat, char* fmt, ... ) {

     va_list ap;
     va_start(ap, fmt);
     if (debug_syn_to_stderr){
       vfprintf(stderr, fmt, ap);
       fputs("\n", stderr);
     }else{
       gpm_report(line,file,stat,fmt,&ap);
     }
     va_end(ap);

   }
#  define gpm_report gpm_report_static
#endif

* END BADE CODE */

/* Prototype */

static void syn_ps2_absolute_mode(int fd);


/* Defines */

#define abs(value)              ((value) < 0 ? -(value) : (value))
#define check_bits(value,mask)  (((value) & (mask)) == (mask))
#define sqr(value)              ((value) * (value))
#define distance(dx,dy)         (sqr(dx) + sqr(dy))
#define mod4(value)             ((value) % 4)
#define max(x,y)                ((x)>(y)?(x):(y))

/*
** Define the edge bit values.
*/
#define LEFT_EDGE           0x01
#define RIGHT_EDGE          0x02
#define TOP_EDGE            0x04
#define BOTTOM_EDGE         0x08
#define UP_BUTTON           0x01
#define DOWN_BUTTON         0x02
#define LEFT_BUTTON         0x04
#define RIGHT_BUTTON        0x08
#define UPPER_LEFT_CORNER   (LEFT_EDGE  | TOP_EDGE)
#define LOWER_LEFT_CORNER   (LEFT_EDGE  | BOTTOM_EDGE)
#define UPPER_RIGHT_CORNER  (RIGHT_EDGE | TOP_EDGE)
#define LOWER_RIGHT_CORNER  (RIGHT_EDGE | BOTTOM_EDGE)
#define is_corner(edges)    (((edges) & (LEFT_EDGE | RIGHT_EDGE)) && \
			     ((edges) & (TOP_EDGE  | BOTTOM_EDGE)))


/****************************************************************************
**
** Configuration constants.
**
** Those with C at the end are suitable for adjustment by a configuration
** program.
****************************************************************************/

static int corner_taps_enabled         = 1;  /* are corner taps enabled C*/
static int tap_gesture_enabled         = 1;  /* are gestures treaded as taps enabled  C*/
static int tossing_enabled             = 1;  /* is toss/catch enabled   C*/
static int does_toss_use_static_speed  = 1;  /* is toss/catch speed     C*/
                                                /* based on toss dist       */
static int edge_motion_enabled         = 1;  /* is edge motion enabled  C*/
static int edge_motion_speed_enabled   = 1;  /* does pressure control   C*/
                                               /* speed of edge motion     */
static int pressure_speed_enabled      = 1;  /* does pressure control   C*/
                                                 /* speed in non edges       */
static int tap_hold_edge_motion_enabled = 1; /* Enable edge motion while holding a tap  C*/

/* pressure induced speed related configuration constants */
static int   low_speed_pressure      = 60;                                /*C*/
static int   speed_up_pressure       = 60;                                /*C*/
static float speed_pressure_factor   = 0.05;                              /*C*/
static float standard_speed_factor   = 0.08;                              /*C*/

/* toss/catch related constants */
static int   min_toss_time           = 100;   /* ms: 0.10 sec */
static int   max_toss_time           = 300;   /* ms: 0.30 sec */          /*C*/
static int   prevent_toss_time       = 300;   /* ms: 0.25 sec */          /*C*/
static int   min_toss_dist           = 2;     /* mm */                    /*C*/
static int   static_toss_speed       = 70;                                /*C*/
static float toss_speed_factor       = 0.5;                               /*C*/

/* edge motion related configuration constants */
static int   x_min_center            = 1632;  /* define left edge           C*/
static int   x_max_center            = 5312;  /* define right edge          C*/
static int   y_min_center            = 1408;  /* define bottom edge         C*/
static int   y_max_center            = 4108;  /* define top edge            C*/
static int   edge_speed              = 20;    /* default speed at edges     C*/

/* gesture related configuration constants for when wmode is enabled. */
static int   wmode_enabled	     = 1;     /* is wmode enabled        C */
static int   drag_lock_enabled	     = 1;     /* is drag locking enabled C */
static int   finger_threshold	     = 30;    /* pressure before it is a finger C */
static int   tap_lower_limit	     = 5;     /* a tap last at least this long C */
static int   tap_upper_limit	     = 200;   /* a tap is at most this long C */
static int   tap_range		     = 100;   /* mm finger movement limit  C */
static int   tap_interval	     = 200;   /* a tap reports button press this long C*/
static int   multiple_tap_delay      = 30;    /* time between reported button pressed C*/
static int   pads_tap_interval       = 8;     /* if pad sends gestures, what is it's tap_interval. C*/

/* wmode capabilities */
static int   palm_detect_enabled     = 1;     /* Ignore when palm on pad   C */
static int   palm_detect_level       = 12;    /* Detecting a palm level (between 0-11) C */
static int   multi_finger_tap_enabled= 1;     /* No of fingers decides which button is pressed C*/
static int   multi_finger_stop_enabled = 1;   /* less sensitive mouse with multi finger C*/
static int   multi_finger_stop_delay = 8;     /* how long to stop after a multifinger detection. C*/


/* mixed configurations */
static int   touchpad_enabled        = 1; /* Disable the touch pad, see corner action turn_on_off. C*/
static int   stick_enabled           = 1; /* Some machines have a touchpad and a stick on
                                           * same device port. The stick will be ignored in
                                           * absolute mode, this option try to recognize
                                           * it's packets.                                         C*/
static int   stick_pressure_enabled  = 0; /* A (hard) press on the stick is reported as left click C*/
static int   four_way_button_enabled = 1; /* Round button giving 4 choices on some touchpads        C*/
static int   four_way_button_is_mouse= 1; /* Is the button: 4 buttons or does it moves the mouse.  C*/
static int   scrolling_enabled       = 1; /* Simulate wheel mouse in at the right edge             C*/
static int   scrolling_edge          = RIGHT_EDGE; /* Which edge is a the scrolling edge           C*/
static int   scrolling_speed         = 10;/* less is faster, 1 fastest  */
static float scrolling_button_factor = 0.5; /* How fast should a button/corner tap scroll, higher faster C*/
static int   auto_scrolling_enabled  = 1; /* Moving to the upper/lower edge keeps scrolling up/downC*/
static int   auto_scrolling_factor   = 2.0; /* How fast should autoscrolling be                   C*/
static int   reset_on_error_enabled  = 0; /* If a packet does not conform to any absolute protocol 
					   * should we reset the touchpad? This is wrong, because we
					   * should rather find out why it does that in first place.
					   * Do not turn it on per default.                         */


/*
** Types for describing actions.  When adding a new action, place it as
** the last item. It will break old configurefiles otherwise.
*/
typedef enum {
  No_Action          = 0,
  Left_Button_Action,
  Middle_Button_Action,
  Right_Button_Action,
  Fourth_Button_Action,
  Up_Button_Action,
  Down_Button_Action,
  Turn_On_Off_Action,
  Debug_On_Off_Action,
  Reset_Touchpad_Action,
  Toggle_Four_Way_Button_Action,
  Toggle_Stick_Pressure_Action,
  Toggle_Scrolling_Action
} action_type;


typedef struct {
  int         action_mask;
  action_type action;
} touchpad_action_type;

static touchpad_action_type corner_actions [] = {
  { UPPER_LEFT_CORNER,  No_Action },
  { LOWER_LEFT_CORNER,  No_Action },
  { UPPER_RIGHT_CORNER, Middle_Button_Action },
  { LOWER_RIGHT_CORNER, Right_Button_Action  },
  { 0,                  No_Action } /* stop flag value */
};


static touchpad_action_type four_button_action [] = {
  { UP_BUTTON,          Up_Button_Action },
  { DOWN_BUTTON,        Down_Button_Action },
  { LEFT_BUTTON,        Middle_Button_Action },
  { RIGHT_BUTTON,       Fourth_Button_Action },
  { 0,                  No_Action } /* stop flag value */
};

/*
** These types are used to read the configuration data from the config file.
*/
typedef enum {
  Integer_Param,
  Float_Param,
  Flag_Param,
  Action_Param
} param_type_type;

typedef struct {
  char            *name;
  param_type_type  p_type;
  union {
    void  *gen_p;   /* avoids complaints by the compiler later on */
    int   *int_p;
    float *float_p;
    int   *flag_p;
    touchpad_action_type *corner_p;
  } addr;
} param_data_type;

static param_data_type param_data [] = {
  /* enabling configuration parameters */
  { "edge_motion_enabled",        Flag_Param,    {&edge_motion_enabled        }},
  { "edge_motion_speed_enabled",  Flag_Param,    {&edge_motion_speed_enabled  }},
  { "corner_taps_enabled",        Flag_Param,    {&corner_taps_enabled        }},
  { "tap_gesture_enabled",        Flag_Param,    {&tap_gesture_enabled        }},
  { "pressure_speed_enabled",     Flag_Param,    {&pressure_speed_enabled     }},
  { "tossing_enabled",            Flag_Param,    {&tossing_enabled            }},
  { "does_toss_use_static_speed", Flag_Param,    {&does_toss_use_static_speed }},
  { "tap_hold_edge_motion_enabled",Flag_Param,   {&tap_hold_edge_motion_enabled}},
  
  /* pressure induced speed related configuration parameters */		     
  { "low_pressure",               Integer_Param, {&low_speed_pressure         }},
  { "speed_up_pressure",          Integer_Param, {&speed_up_pressure          }},
  { "pressure_factor",            Float_Param,   {&speed_pressure_factor      }},
  { "standard_speed_factor",      Float_Param,   {&standard_speed_factor      }},
  /* toss/catch related parameters */		 			     
  { "min_toss_time",              Integer_Param, {&min_toss_time              }},
  { "max_toss_time",              Integer_Param, {&max_toss_time              }},
  { "prevent_toss_time",          Integer_Param, {&prevent_toss_time          }},
  { "min_toss_dist",              Integer_Param, {&min_toss_dist              }},
  { "static_toss_speed",          Integer_Param, {&static_toss_speed          }},
  { "toss_speed_factor",          Float_Param,   {&toss_speed_factor          }},
  /* edge motion related configuration parameters */			     
  { "x_min_center",               Integer_Param, {&x_min_center               }},
  { "x_max_center",               Integer_Param, {&x_max_center               }},
  { "y_min_center",               Integer_Param, {&y_min_center               }},
  { "y_max_center",               Integer_Param, {&y_max_center               }},
  { "edge_speed",                 Integer_Param, {&edge_speed                 }},
  /* corner tap actions */			 			     
  { "upper_left_action",          Action_Param,  {&corner_actions [0]         }},
  { "lower_left_action",          Action_Param,  {&corner_actions [1]         }},
  { "upper_right_action",         Action_Param,  {&corner_actions [2]         }},
  { "lower_right_action",         Action_Param,  {&corner_actions [3]         }},
  /* 4 way button actions */
  { "up_button_action",           Action_Param,  {&four_button_action [0]     }},
  { "down_button_action",         Action_Param,  {&four_button_action [1]     }},
  { "left_button_action",         Action_Param,  {&four_button_action [2]     }},
  { "right_button_action",        Action_Param,  {&four_button_action [3]     }},
  /* use wmode */				 			     
  { "wmode_enabled",		  Flag_Param,	 {&wmode_enabled	      }},
  { "drag_lock_enabled",	  Flag_Param,	 {&drag_lock_enabled	      }},
  { "finger_threshold",		  Integer_Param, {&finger_threshold	      }},
  { "tap_lower_limit",		  Integer_Param, {&tap_lower_limit	      }},
  { "tap_upper_limit",		  Integer_Param, {&tap_upper_limit	      }},
  { "tap_range",		  Integer_Param, {&tap_range		      }},
  { "tap_interval",		  Integer_Param, {&tap_interval		      }},
  { "multiple_tap_delay",	  Integer_Param, {&multiple_tap_delay	      }},
  { "pads_tap_interval",	  Integer_Param, {&pads_tap_interval	      }},
  /* Additional wmode parameters */				 			     
  { "palm_detect_enabled",	  Flag_Param,	 {&palm_detect_enabled	      }},
  { "palm_detect_level",	  Integer_Param, {&palm_detect_level	      }},
  { "multi_finger_tap_enable",    Flag_Param,    {&multi_finger_tap_enabled   }},
  { "multi_finger_stop_enabled",  Flag_Param,    {&multi_finger_stop_enabled  }},
  { "multi_finger_stop_delay",    Integer_Param, {&multi_finger_stop_delay    }},
  /* Additional options*/
  { "touchpad_enabled", 	  Flag_Param,    {&touchpad_enabled	      }},
  { "stick_enabled",	          Flag_Param,    {&stick_enabled	      }},
  { "stick_pressure_enabled",     Flag_Param,    {&stick_pressure_enabled     }},
  { "four_way_button_enabled",    Flag_Param,    {&four_way_button_enabled    }},
  { "four_way_button_is_mouse",   Flag_Param,    {&four_way_button_is_mouse   }},
  { "scrolling_enabled",	  Flag_Param,    {&scrolling_enabled	      }},
  { "auto_scrolling_enabled",	  Flag_Param,    {&auto_scrolling_enabled     }},
  { "scrolling_edge",	          Integer_Param, {&scrolling_edge	      }},
  { "scrolling_speed",	          Integer_Param, {&scrolling_speed	      }},
  { "scrolling_button_factor",    Float_Param,   {&scrolling_button_factor    }},
  { "auto_scrolling_factor",      Float_Param,   {&auto_scrolling_factor      }},
  /* end of list */				 			     
  { NULL,                         Flag_Param,    {NULL                        }}
};


/*
** The information returned in the identification packet.
** STIG page 10
*/
typedef struct {
  int info_model_code;
  int info_major;
  int info_minor;
} info_type;

static info_type ident;

/*
** The information returned in the model ID packet.
** STIG page 11
*/

#define INFO_ROT180_BITS      0x800000   /* bit 23    */
#define INFO_PORTRAIT_BITS    0x400000	 /* bit 22    */
#define INFO_SENSOR_BITS      0x3F0000	 /* bit 16-21 */
#define INFO_HARDWARE_BITS    0x00FE00	 /* bit 9-15  */
#define INFO_NEW_ABS_BITS     0x000080	 /* bit 7     */
#define INFO_CAP_PEN_BITS     0x000040	 /* bit 6     */
#define INFO_SIMPLE_CMD_BITS  0x000020	 /* bit 5     */
#define INFO_GEOMETRY_BITS    0x00000F	 /* bit 0-3   */


typedef struct {
  int info_rot180; 
  int info_portrait; 
  int info_sensor; 
  int info_hardware; 
  int info_new_abs; 
  int info_cap_pen; 
  int info_simple_cmd; 
  int info_geometry; 
} model_id_type; 

static model_id_type model;

/*
** The sensor types as of STIG 2.5
** Page 11. Plus some guessing.
*/
static char *model_names [] = {
  "Unknown",                                /*  0 */
  "Standard TouchPad (TM41xx134)",
  "Mini Module (TM41xx156)",
  "Super Module (TM41xx180)",
  "Romulan Module",
  "Apple Module",
  "Single Chip",
  "Flexible pad (discontinued)",
  "Ultra-thin Module (TM41xx220)",
  "Wide pad Module (TW41xx230)",
  "TwinPad",                                /* 10 */
  "StampPad Module (TM41xx240)",
  "Sub-mini Module (TM41xx140)",
  "MultiSwitch module (TBD)",
  "Standard Thin",
  "Advanced Technology Pad (TM41xx221)",
  "Ultra-thin Module, connector reversed",
  "Low Power Module",
  "Thin Module, ATP",
  "Snap dome Module, ATP", 
  "FlexArm Module",                         /* 20 */
  "TWIII Module (TouchWriter 3)",
  "TWIII Module (P64)",
  "Combo Module",
  "Squish Module",
  "Thin TTL Serial module",
  "TWIII Ultra Thin Module",
  "PS/2 Passthrough", 
  "4 Button On Board Module",
  "6 Buttons Off Board Module",             
  "6 Buttons On Board Module",              /* 30 */
  "Ultrathin TTL serial Module",
  "ClearPad Module",
  "HyperThin Sensor Module",
  "Pad With Scroll Strip",
  "UltraThin TTL rounded, serial Module",
  "Ultrathin ATP Module",
  "SubMini w/6 buttons",
  "Standard USB Module",
  "cPad Dropin Plain USB Module",
  "cPad"                                    /* 40 */ 
};





/*
** Define the information known about a sensor.
** STIG page 14.
** Resolution only apply for absolute mode.
** For older models the default resolution is 85x94.
*/
typedef struct {
  char *model;
  int   x_per_mm;
  int   y_per_mm;
  float width_mm;
  float height_mm;
} sensor_info_type;


static sensor_info_type sensor_info [] = {
  { "",            0,   0,  0.0,  0.0 },
  { "Standard",   85,  94, 47.1, 32.3 },
  { "Mini",       91, 124, 44.0, 24.5 },
  { "Super",      57,  58, 70.2, 52.4 },
  { "",            0,   0,  0.0,  0.0 },
  { "",            0,   0,  0.0,  0.0 },
  { "",            0,   0,  0.0,  0.0 },
  { "",            0,   0,  0.0,  0.0 },
  { "UltraThin",  85,  94, 47.1, 32.3 },
  { "Wide",       73,  96, 54.8, 31.7 },
  { "",            0,   0,  0.0,  0.0 },
  { "Stamp",     187, 170, 21.4, 17.9 },
  { "SubMini",   122, 167, 32.8, 18.2 },
};

static sensor_info_type *sensor;

/*
** The information returned in the extended capability packet.
** STIG page 15
*/

#define EXT_CAP_EXTENDED     0x8000     /* Bit 15  */
#define EXT_CAP_STICK        0x0080     /* Bit 8   */
#define EXT_CAP_SLEEP        0x0010     /* Bit 4   */
#define EXT_CAP_FOUR_BUTTON  0x0008     /* Bit 3   */
#define EXT_CAP_MULTI_FINGER 0X0002     /* Bit 1   */
#define EXT_CAP_PALM_DETECT  0X0001     /* Bit 0   */

typedef struct {
  int cap_ext; 
  int cap_stick;
  int cap_sleep; 
  int cap_four_button; 
  int cap_multi_finger; 
  int cap_palm_detect; 
} ext_cap_type;

static ext_cap_type capabilities;

/*
** The information in the mode byte.
** STIG Page 17
*/
#define ABSOLUTE_MODE      0x80      /* Bit 7 set */
#define RELATIVE_MODE      0x00	     /* Bit       */
#define HIGH_REPORT_RATE   0x40	     /* Bit 6 set = 80 packages per second */
#define LOW_REPORT_RATE    0x00	     /* Bit         40 packages per second */
#define USE_9600_BAUD      0x08	     /* Bit 3 for serial protocol */
#define USE_1200_BAUD      0x00	     /* Bit   */
#define PS2_SLEEP          0x08	     /* Bit 3 for ps2 protocol */
#define PS2_NO_SLEEP       0x00	     /* Bit   (only  button press activates touchpad again)*/
#define NO_TAPDRAG_GESTURE 0x04      /* Bit 2 for model version >= 4 */
#define TAPDRAG_GESTURE    0x00      /* Bit   */     
#define EXTENDED_REPORT    0x02	     /* Bit 1 for serial protocol absolute mode only */
#define NORMAL_REPORT      0x00	     /* Bit   */
#define STICK_DISABLE      0x02	     /* Bit 1 for ps2 protocol this disables any stick attached */
#define STICK_ENABLED      0x00	     /* Bit   */
#define REPORT_W_ON        0x01	     /* Bit 0 set */
#define REPORT_W_OFF       0x00      /* Bit   */



/*
** The format of the translated data to a report.
*/
typedef struct {
  int left;
  int middle;
  int right;
  int fourth;
  int up;
  int down;
  int x;
  int y;
  int pressure;
  int gesture;
  int fingers;
  int w;
} report_type;

static report_type last_report,report;

/*
** A location record.
** This is needed to make an average over several packages, 
** because the reported x,y might not be that accurate.
*/
typedef struct {
  int x;
  int y;
} location_type;



/*
** Parameters for controlling the touchpad.
*/
/* touchpad information */
static int res_x;
static int res_y;
static int x_per_mm;
static int y_per_mm;


/* status information */
static int           packet_num = 0;
static int           was_edges = 0;
static int           was_non_edge = 0;
static int           last_corner_action = GPM_B_NONE;
static location_type last_locs [4];
static Gpm_Event     last_state;
static int           tap_lower_limit_packet;    
static int           tap_upper_limit_packet;

/* toss status information */
static int           is_tossing = 0;
static int           was_tossing = 0;
static int           min_toss_dist__2;
static int           max_toss_packets;
static int           min_toss_packets;
static int           prevent_toss_packets;
static int           toss_timer;
static location_type toss_speed;
static location_type touch_loc;


/* Multi finger information */
static int           was_fingers = 0;
static int           multi_finger_stop_timer = 0;
static int           multi_finger_reset_timer;
static int           multi_finger_pressure  = 0;
static int           multi_finger_xy = 0;



/* Scrolling information */
static int           is_scrolling = 0;            /* Scrolling using touchpad edge*/
static int           is_always_scrolling = 0;     /* Only report scrolling, no mouse movement */
static int           scrolling_speed_timer = 0;
static int           scrolling_amount_left = 0;   /* Tells how much to scroll up or down */





/****************************************************************************
**
** ROUTINES for processing either type of touchpad.
**
****************************************************************************/

/*
** Dump the report data for debugging.
**
** Because the synaptics sends (trivial) data in one second after last touch,
** which makes reading the debug data harder, only dump the report if it is different 
** than the previously dumped.
*/
static void tp_dump_report_data (report_type report,
				 int edges,
				 Gpm_Event* state) 
{
  static report_type  last_report_reported;
  static unsigned int times_report_repeated = 0;


  report.left   |= state->buttons & GPM_B_LEFT;
  report.right  |= state->buttons & GPM_B_RIGHT;
  report.middle |= state->buttons & GPM_B_MIDDLE;

  if (memcmp(&report,&last_report_reported,sizeof(report_type)) == 0){
    times_report_repeated++;
    return;
  }

  /* Was the last report repeated ? */
  if(times_report_repeated > 0){
    gpm_report (GPM_PR_DEBUG,"\rSynps2: Last report reported %d times\n",times_report_repeated); 
    times_report_repeated = 0;
  }

  last_report_reported = report;

  gpm_report (GPM_PR_DEBUG,
	      "\rSynps2: %c%c%c%c%c  %4dx%-4d %3d %2d %d  %c%c%c%c  %c%c    %3d%3d %d    %8d %8d %c",
	      report.fingers ? 'f' : '-',  
	      report.gesture ? 'g' : '-',
	      
	      report.left    ? 'l' : '-',
	      report.middle  ? 'm' : '-',  
	      report.right   ? 'r' : '-',
	      
	      report.x, report.y, report.pressure,
	      report.w,was_fingers, 
	      
	      edges & LEFT_EDGE   ? 'l' : '-',
	      edges & RIGHT_EDGE  ? 'r' : '-',
	      edges & BOTTOM_EDGE ? 'b' : '-',
	      edges & TOP_EDGE    ? 't' : '-',
	      
	      report.gesture && !report.fingers ? 't' : '-',
	      report.gesture && report.fingers  ? 'd' : '-',
	      
	      state->dx,state->dy,state->buttons,


	      multi_finger_pressure,multi_finger_xy,
	      (multi_finger_pressure>4500 && multi_finger_xy>50000? 'f':' '));

}



/* Get model name, STIG page 11 */
static char *syn_model_name (int sensor) 
{
  if (sensor < 0 || 40 < sensor ) {
    return "Reserved";
  } else {
    return model_names [sensor];
  }
}

/* convert the model id from bits to values 
*  STIG page 11 
*/
static void syn_extract_model_id_info (int model_int, model_id_type *model) 
{
  model->info_rot180     = check_bits (model_int,  INFO_ROT180_BITS);
  model->info_portrait   = check_bits (model_int,  INFO_PORTRAIT_BITS);
  model->info_sensor     =            (model_int & INFO_SENSOR_BITS)   >> 16;
  model->info_hardware   =            (model_int & INFO_HARDWARE_BITS) >> 8;
  model->info_new_abs    = check_bits (model_int,  INFO_NEW_ABS_BITS);
  model->info_cap_pen    = check_bits (model_int,  INFO_CAP_PEN_BITS);
  model->info_simple_cmd = check_bits (model_int,  INFO_SIMPLE_CMD_BITS);
  model->info_geometry   =            (model_int & INFO_GEOMETRY_BITS);
}


/* Translate the reported data into a record for processing 
 *  STIG page 14*/
static sensor_info_type *syn_get_sensor_info (int sensor_id) 
{
  if (sensor_id < 0 || 12 < sensor_id ) {
    return &sensor_info [0];
  } else {
    return &sensor_info [sensor_id];
  }
}



/* Translate the reported data of extended capabilities STIG page 15.  If the
 extended bit is not set it should be assumed that neither of the other
 capabilities is available.*/
static void syn_extract_extended_capabilities(int ext_cap_int, ext_cap_type *cap)
{

# ifdef DEBUG  
  gpm_report(GPM_PR_INFO,"Synaptics Touchpad Capabilities: %02X",ext_cap_int);
# endif 

  cap->cap_ext            = check_bits (ext_cap_int, EXT_CAP_EXTENDED);

  if(cap->cap_ext){
    cap->cap_stick        = check_bits (ext_cap_int, EXT_CAP_STICK);
    cap->cap_sleep        = check_bits (ext_cap_int, EXT_CAP_SLEEP);
    cap->cap_four_button  = check_bits (ext_cap_int, EXT_CAP_FOUR_BUTTON);
    cap->cap_multi_finger = check_bits (ext_cap_int, EXT_CAP_MULTI_FINGER);
    cap->cap_palm_detect  = check_bits (ext_cap_int, EXT_CAP_PALM_DETECT);
  }else{
    cap->cap_stick        = 0;
    cap->cap_sleep        = 0;
    cap->cap_four_button  = 0;
    cap->cap_multi_finger = 0;
    cap->cap_palm_detect  = 0;
    wmode_enabled         = 0; /* neither is wmode support by the pad */
  }

  if(!cap->cap_palm_detect)
    palm_detect_enabled   = 0;
}


/*
** Check for edges.
*/
static int tp_edges (location_type loc) 
{
  int edges = 0;
 
  if (loc.x <= x_min_center)
    edges |= LEFT_EDGE;

  if (loc.x >= x_max_center)
    edges |= RIGHT_EDGE;

  if (loc.y <= y_min_center)
    edges |= BOTTOM_EDGE;

  if (loc.y >= y_max_center)
    edges |= TOP_EDGE;

  return edges;
}


/**
 ** Handle scrolling. The wheel is way too fast to be usefull, so only report
 ** every scrolling_speed. This function is also called when scrolling is done by
 ** buttons, which is why DOWN and UP must be removed in the case of not scrolling.
 */

static void tp_handle_scrolling(Gpm_Event *state)
{
  /* Limit the amount of scrolling, so that we do not overrun. */
  if ( scrolling_amount_left >  256*20 )
    scrolling_amount_left = 256*20;

  if ( scrolling_amount_left < -256*20 )
    scrolling_amount_left = -256*20;

  
  if (scrolling_amount_left > scrolling_speed){
    scrolling_amount_left -= scrolling_speed;
    state->buttons |= GPM_B_DOWN;
  }else if (scrolling_amount_left < -scrolling_speed){
    scrolling_amount_left += scrolling_speed;
    state->buttons |= GPM_B_UP;
  }else {
    state->buttons &= ~(GPM_B_DOWN | GPM_B_UP);
  }
    
}


/*
** process_action
**
** Do the action and return a button state for a given action list and mask.
** Actions that should not be repeated should report GPM_B_NONE (when holding
** down a button).
*/

static int tp_process_action(touchpad_action_type *action_list, int mask)
{
  int i = 0;
  int status = GPM_B_NONE;

  if (mask == 0){
    return GPM_B_NONE;
  }

  while (action_list [i].action_mask) {
    if (check_bits (mask, action_list [i].action_mask)) {
      switch (action_list [i].action) {
	case Left_Button_Action:
	status |= GPM_B_LEFT;
	  break;
	case Middle_Button_Action:
	status |= GPM_B_MIDDLE;
	  break;
	case Right_Button_Action:
	status |= GPM_B_RIGHT;
	  break;
	case Fourth_Button_Action:
	status |= GPM_B_FOURTH;
	  break;
        case Up_Button_Action:
	scrolling_amount_left += scrolling_button_factor * scrolling_speed;
	status = GPM_B_UP;
          break;
        case Down_Button_Action:
	scrolling_amount_left -= scrolling_button_factor * scrolling_speed;
	status = GPM_B_DOWN;
          break;
	case Turn_On_Off_Action:
	touchpad_enabled = !touchpad_enabled;
	status = GPM_B_NONE;
	  break;
	case Debug_On_Off_Action:
#         ifdef DEBUG_SYNAPTIC
	    debug_syn_to_stderr = !debug_syn_to_stderr;
#         endif
	status = GPM_B_NONE;
	  break;
	case Reset_Touchpad_Action:
	  syn_ps2_reset(which_mouse->fd);
	  syn_ps2_absolute_mode(which_mouse->fd);
	status = GPM_B_NONE;
	  break;
	case Toggle_Four_Way_Button_Action:
	  four_way_button_is_mouse = !four_way_button_is_mouse;
	status = GPM_B_NONE;
	break;
      case Toggle_Stick_Pressure_Action:
	stick_pressure_enabled = !stick_pressure_enabled;
	status = GPM_B_NONE;
	  break;
      case Toggle_Scrolling_Action:
	is_always_scrolling = !is_always_scrolling;
	status = GPM_B_NONE;
	case No_Action:
	default:
	status = GPM_B_NONE;
	  break;
	}
      }

      i++;
    }

  return status;
}


/*
** process_corner_taps. 
**
** last_corner_action holds the last action because then the second "tap" in
** Tap hold and two taps do not have to be at the corner, we just repeat the
** action. This function is only called the at the first gesture report; at
** consecutive gestures, the buttons are repeated other place. Action that
** should not be repeated should report GPM_B_NONE.
*/
static void tp_process_corner_taps (Gpm_Event *state, report_type report) 
{
  if (corner_taps_enabled && report.gesture &&
      (is_corner (was_edges) || last_corner_action != GPM_B_NONE)) {

    gpm_report (GPM_PR_DEBUG,"\rTouchpad: Corner action"); 
    last_corner_action = tp_process_action(corner_actions,was_edges);

    state->buttons |= last_corner_action;
  } else {
    last_corner_action = GPM_B_NONE;
  }
}



/*
** syn_ps2_process_extended_packets
**
** The internal synaptics ps2 touchpads can have extra devices attached (stick,
** 4 way button) to them, and they reports there state as special embedded
** packets. These attached devices are handled here.
**
** The function returns 1 if it does not make sense to continue normal touchpad
** processing, 0 otherwise.
**            
** I dont think that the external touchpads (serial) can have the same devices
** attached.  This code should be moved to ps2 specific part, but it is easy to
** have it here.
*/

static int syn_ps2_process_extended_packets( unsigned char *data, 
				   report_type *report,
					     Gpm_Event *state)
{

  /* Sanity check of data. */
  if (report->pressure == 0 && (report->x != 0 || report->y != 0) ){

     /* Something is wrong, should we assume it is an extended packet?  It
     *  cannot be processed further than here as the pressure is 0, which would
     *  break things, if the user uses it simultaneously with the touchpad. 
     */
    /* Allow some simultaneously uses: tap hold on touchpad, then extended movement. */
    state->buttons |= last_state.buttons;

    /* Stick invariant bits (I hope). See absolute packets */
    /* Stick pressed: The stick do only generates one packet, so double tap is
     * a problem; squeezing in a non-clicked state. It is probably not a real
     * problem, as it is hard not to move the stick a little between pressing,
     * thereby returning a non-clicked state between the packets. Maybe
     * press-lock mechanism is useful. Forget it, the styk supports it! */
    if((data[0] & 0xFC) == 0x84 && 
       (data[1] & 0xCE) == 0x08 &&
       (data[3] & 0xFC) == 0xC4){

      if (stick_enabled){
	
	state->buttons |=
	  ((data[0]& 0x01) ? GPM_B_LEFT  : 0 )+
	  ((data[0]& 0x02) ? GPM_B_RIGHT : 0 )+
	  (((data[0] ^ data[3]) & 0x01) ? GPM_B_MIDDLE : 0) +
	  (((data[0] ^ data[3]) & 0x02) ? GPM_B_FOURTH : 0);
	
	state->dx=  ((data[1] & 0x10) ? data[4]-256 : data[4]);
	state->dy= -((data[1] & 0x20) ? data[5]-256 : data[5]);
      }

      if (stick_pressure_enabled && 
	  data[1] & 0x01){
	state->buttons |= GPM_B_LEFT;
      }
            
#     ifdef DEBUG_STICK      
      gpm_report (GPM_PR_DEBUG,"StickData? %02x %02x %02x %02x %02x %02x :dx:%d dy:%d b:%d",
		  data[0],data[1],data[2],data[3],data[4],data[5],
		  state->dx,state->dy,state->buttons);
#     endif
      return 1;
    }

    /* 4 way button invariant bits (I hope). I dont know the official name. */
    if((data[0] & 0xFC) == 0x80 &&
       (data[1] & 0xFF) == 0x00 &&
       (data[3] & 0xFC) == 0xC0 &&
       report->x < 4 && report->y < 4 ){

      if(four_way_button_enabled){

	state->buttons|=
	  ((data[0]& 0x01) ? GPM_B_LEFT  : 0 )+
	  ((data[0]& 0x02) ? GPM_B_RIGHT : 0 )+
	  (((data[0] ^ data[3]) & 0x01) ? GPM_B_MIDDLE : 0) +
	  (((data[0] ^ data[3]) & 0x02) ? 0 : GPM_B_FOURTH); /* Note that this is reversed. */


	if (four_way_button_is_mouse){
	  /* Report motion */
	  if (report->x & 1) /* UP */
	    state->dy = -1;
	  if (report->y & 1) /* DOWN */
	    state->dy = 1;
	  if (report->x & 2) /* LEFT */
	    state->dx = -1;
	  if (report->y & 2) /* RIGHT */
	    state->dx = 1;
	}else{ 
	  /* Report buttons */
	  if (report->x & 1) /* UP */
	    state->buttons |= tp_process_action(four_button_action,UP_BUTTON);
	  if (report->y & 1) /* DOWN */
	    state->buttons |= tp_process_action(four_button_action,DOWN_BUTTON);
	  if (report->x & 2) /* LEFT */
	    state->buttons |= tp_process_action(four_button_action,LEFT_BUTTON);
	  if (report->y & 2) /* RIGHT */
	    state->buttons |= tp_process_action(four_button_action,RIGHT_BUTTON);
	}
      }

      if ( scrolling_amount_left != 0 ){
	tp_handle_scrolling(state);
      }

      return 1;
    } 

    /* This is unknown packet. */
    gpm_report (GPM_PR_ERR,"\rSynps2: Pressure is 0, but x or y is not 0. "
		           "Data: %02X %02X %02X %02X %02X %02X",
		data [0],data [1],data [2],data [3],data [4],data [5]);
    return 1;
  }

  return 0;
}

/**
** Preprocess the report even before doing wmode stuff.
** Is always called directly after the conversion to check the data received.
** return 0 if it is reasonable, 1 if there is something wrong.
**
** Checks for correct data package, palm on the pad, number of fingers.
**            
*/

static int tp_preprocess_report ( report_type *report,
				  Gpm_Event *state)
{

  static int fake_extra_finger;
  static int was_fake_pressure;

  /* Check whether there is a palm on the pad */
  /* STIG page 9: Values of w, vary from pad to pad. It is not precise when pressure is small < 25.
   * 4-7  finger of normal width
   * 8-14 very wide finger or palm
   * 15   maximum reportable width     
   */
  if (wmode_enabled && palm_detect_enabled &&  
      report->pressure != 0 && ((report->w - 4) >= palm_detect_level)){
#   ifdef DEBUG_PALM
    gpm_report (GPM_PR_DEBUG,"\rSynps2: palm detected. w: %d",report->w); 
#   endif
    /* BUG should not return 1, as this drops packets. Return a repeated report ? */
    /*       last_locs [mod4 (packet_num - 1)].x = report->x; */
    /*       last_locs [mod4 (packet_num - 1)].y = report->y; */
    /*     or */
    /*       report->pressure = 0; */
    return 1;
  }


    
  /* Check whether there is one finger on the pad,
  *  but only if pad does not report it, as it also report gestures then,
  *  which it would signal independed on whether we detect fingers. */
  if(wmode_enabled){
    report->fingers  = (report->pressure > finger_threshold);
  }
  
  if (multi_finger_tap_enabled || multi_finger_stop_enabled){
    
    if (wmode_enabled && capabilities.cap_multi_finger){
      
      /* use w values. w = 0: 2 fingers, w = 1: 3 fingers, (if there is pressure) */
      if (report->pressure != 0 && (report->w == 0 || report->w == 1)){
	report->fingers = 2+report->w;
	fake_extra_finger = 0;
      }
    }

    /* Extra check for vertical multi fingers which my pad is very bad to detect. 
     * Only check for extra fingers if w has not changed.
     * Faking fingers may go wrong so sanity check: */
    if(report->fingers == 0){
      fake_extra_finger = 0;
#     ifdef DEBUG_REPORTS
        multi_finger_pressure  = 0;
        multi_finger_xy = 0;
#     endif
    }else
      if (report->fingers + fake_extra_finger == last_report.fingers){

	multi_finger_pressure = sqr(report->pressure) - sqr(last_report.pressure);
	multi_finger_xy = (sqr(last_locs [mod4 (packet_num - 1)].x - report->x) +
			   sqr(last_locs [mod4 (packet_num - 1)].y - report->y) );

	/* Check for a second finger. If the move is larger than tap range,
	* then it is not a tap, and adding the finger does not change reported        
	* buttons, but it will still stop the moving if
	* multi_finger_stop_enabled is on. These tests are complete base on my
	* imagination and experience, so any better idea are welcome. */
	if(report->fingers == 1 && 
	   multi_finger_pressure > 4500 && multi_finger_xy > 2*sqr((double)tap_range)){
	    fake_extra_finger = 1;
	    was_fake_pressure =  report->pressure;

	}/* Check for third finger. */
	else if(last_report.pressure  > 180  && (report->fingers + fake_extra_finger) == 2 &&
		 multi_finger_pressure > 4500 && multi_finger_xy > 2*sqr((double)tap_range)){
	  fake_extra_finger = 2;
	  was_fake_pressure = report->pressure;

	} else if ( (fake_extra_finger > 0) &&
		    (was_fake_pressure - report->pressure > 20)){
	  fake_extra_finger --;

	} else if(multi_finger_pressure < -5000 && multi_finger_xy > 2*sqr((double)tap_range)){
	  /* Probably missed a placed multi finger, as this is one removed! */
	  last_report.fingers ++;
	}
      }
    
    report->fingers += fake_extra_finger;
    
  }


  /* Check whether to reduce sensibility when adding or removing fingers */
  if (multi_finger_stop_enabled){

    /* Is a finger added or removed since last packet? */
    if( (report->fingers     > 1 && report->fingers > last_report.fingers) ||
	(last_report.fingers > 1 && report->fingers < last_report.fingers) ){

      /* Updating the timer right after another added/removed finger,
       * would make the undo moving become redo, so dont.*/
      if(multi_finger_stop_timer != multi_finger_stop_delay - 1)
	multi_finger_stop_timer  = multi_finger_stop_delay;

#     ifdef DEBUG_MULTI_FINGER
      gpm_report (GPM_PR_DEBUG,"%s multi finger %d %d %d",
		  report->fingers > last_report.fingers ? "Add":"Remove",
		  last_report.fingers,report->fingers,fake_extra_finger);
#     endif
      
    } /* Should be tested last, because of undo moving when removing fingers. */
    else if(report->fingers == 0){
	multi_finger_stop_timer = 0;
    }

    /* Stop moving the mouse after a finger adding/removing. */
    if( multi_finger_stop_timer > 0){
      last_locs [mod4 (packet_num - 1)].x = report->x;
      last_locs [mod4 (packet_num - 1)].y = report->y;
      last_locs [mod4 (packet_num - 2)].x = report->x;
      last_locs [mod4 (packet_num - 2)].y = report->y;

      /* Undo the previous move (before detecting of the adding/removing). */
      if (multi_finger_stop_timer == multi_finger_stop_delay){
	last_locs [mod4 (packet_num - 2)].x += last_state.dx * 2 / standard_speed_factor;
	last_locs [mod4 (packet_num - 2)].y -= last_state.dy * 2 / standard_speed_factor;	
      }

      multi_finger_stop_timer --;
    }
  }

  /* If wmode is not used, we do not know all the informations to reset the
   * was_fingers variable, so set a timer to reset the variable when we believe it is time.
   * pads_tap_interval is optimal when set to the touchpads tap_interval.
   */

  if(!wmode_enabled){
    if (report->fingers || report->gesture)
      multi_finger_reset_timer = pads_tap_interval;

    if(multi_finger_reset_timer > 0)
      multi_finger_reset_timer --;
    else
      was_fingers = 0;
  }


  was_fingers = max(was_fingers,report->fingers);

  return 0;
}


/** 
** tp_process_wmode_report
** Process the report from a wmode enabled device.  No gesture calculation is
** done by the device in wmode, so the find tap and drag hold and tap hold gestures.
**
** Tap mechanism: High when touched. (=: assigned, -: decreased, +: increased)
**
**                         _________________________  
**                         | tap_lower/upper_limit |tap_interval|  
**         -----------------                       -----------------
** 
** stroke_x/y         :    =                         
** finger_on_pad_timer: 0000123        +++++++++++++0000000000000000
** time_to_forget_tap : 0000000        000000000000=---------3210000
** gesture_delay      : 0000000  ...   00000000000000000000000000000
** drag_locked        : 0000000        00000000000000000000000000000
** gesture            : 0000000        00000000000011111111111110000
**
** Note:
**  0) When gesture is high a button is reported pressed (usually left button).
**  1) finger_on_pad_timer is increased at most to 1 larger than tap_upper_limit,
**     but then it is not a tap any longer.
**  2) If tap_interval is larger than 1s (80 packets) then time_to_forget_tap never reaches 0,
**     before the touchpads stop sending packets, so gesture is reported until touched again.  
**
**
** Tap hold mechanism: High when touched. (=: assigned, -: decreased, +: increased)
**
**                                       tap_interval
**                         ________________  v ___   ___
**                         |tap_l/u_limit |    |not tap|
**         -----------------              ------       --------    
** 
** stroke_x/y         :    =                   =      
** finger_on_pad_timer: 0000123+       +++00000123++  +++0000
** time_to_forget_tap : 00000000       000=----11111  1110000
** gesture_delay      : 00000000 ...   0000000000000  0000000
** drag_locked        : 00000000       0000000000000  0000000
** gesture            : 00000000       0001111111111  1110000
**
** Note:
**  0) time_to_forget_tap is set to 1 when touching the pad 2 time, so the gesture is stop when released. 
**     (two fast taps would be a problem).
**
**
** Two consecutive taps mechanism: High when touched. (=: assigned, -: decreased, +: increased)
**
**                                       tap_interval  multiple_tap_delay
**                         ________________  v _______________  v   |
**                         |tap_l/u_limit |    |tap_l/u_limit|  
**         -----------------              ------             --------    
** 
** stroke_x/y         :    =                   =      
** finger_on_pad_timer: 0000123+       +++00000123++       ++00000000000000000
** time_to_forget_tap : 00000000       000=----11111       11=---------32100000
** gesture_delay      : 00000000 ...   0000000000000 ...   00=----3210000000000
** drag_locked        : 00000000       0000000000000       00000000000000000000
** gesture            : 00000000       0001111111111       11000000001111100000
**
** Note:
**  0) There is no gesture if gesture_delay is non zero.
**  1) multiple_tap_delay (gesture_delay) should be less than tap_interval (time_to_forget_tap),
**     or else no second tap is reported.
**  2) Three fast taps (shorter than multiple_tap_delay) would reset gesture_delay the second time 
**     and only report 2 taps (the first and last). (likewise with more fast taps). BUG ?
**
**
** Drag lock mechanism: High when touched. (=: assigned, -: decreased, +: increased)
** MISSING
*/


static void tp_process_wmode_report (report_type *report) 
{
  static int finger_on_pad_timer = 0;
  static int time_to_forget_tap = 0;
  static int gesture_delay = 0;
  static int stroke_x;
  static int stroke_y;
  static int drag_locked = 0;


  if (report->fingers > 0) {
	    
    /* finger down for the first time */
    if (finger_on_pad_timer == 0) { 
      stroke_x = report->x;
      stroke_y = report->y;
    }
    
    /* don't want timer to overflow */
    if (finger_on_pad_timer < (tap_upper_limit_packet)) 
      finger_on_pad_timer ++; 
    
    /* dragging and consecutive tap gestures is to end with finger up 
     *  forget fast that there was a tap if this is not a part of a tap.*/
    if (time_to_forget_tap > 0) 
      time_to_forget_tap = 1;
    
  } else { /* interesting things happen when finger is up */
      
    /* tap determination: Was the finger long enough on the pad and not too
     * long, while staying at the same place.
     */
    if ((finger_on_pad_timer > (tap_lower_limit_packet)) &&  /* minimum finger down time */
	(finger_on_pad_timer < (tap_upper_limit_packet)) &&  /* maximum finger down time */
	((distance((double)(stroke_x - report->x),  /* maximum range for finger to drift while down */
		  (double)(stroke_y - report->y))
	 < sqr((double)tap_range)) ||
	 (multi_finger_tap_enabled && was_fingers > 1))) {
      
      /* not a consecutive tap? */
      if (time_to_forget_tap == 0) 
	gesture_delay = 0; /* right -> don't delay gesture */
      else { /* a consecutive tap! */
	gesture_delay = multiple_tap_delay * 80 / 1000; /* delay gesture to create multiple click */
      }
      
      /* is drag locked */
      if (drag_locked) {
	drag_locked = 0; /* unlock it and don't gesture. */
	time_to_forget_tap = 0;
      } else 
	time_to_forget_tap = tap_interval * 80 / 1000; /* setup gesture time to count down */
      
    } else { /* It was not a tap */
      
      /* a drag to lock?  If user did a tap and quickly hold the finger longer than a tap.
       */
      if (drag_lock_enabled && 
	  (time_to_forget_tap > 0) && (finger_on_pad_timer >= (tap_upper_limit_packet)))
	drag_locked = 1;
      
      if (time_to_forget_tap  > 0) time_to_forget_tap --;
      if (time_to_forget_tap == 0) was_fingers=0;
      if (gesture_delay > 0) gesture_delay --;
      
    }
    
    finger_on_pad_timer = 0;
    
  }
  
  report->gesture  = ((time_to_forget_tap > 0) && (gesture_delay == 0)) || drag_locked;
}


/*
** tp_process_report
**
** Process the touchpad report. Do tossing mechanism, edge mechanism (edge
** extension, corner taps), speed pressure.
**
** Tossing mechanism: High when touched.  (=: assigned, -: decreased)
**                                                                         (Pause)
**                 ________________________        _____         ______ prevent_toss ___
**                 | min/max_toss_packets |  < 1s  |      ....        |      ...     |  
** -----------------                      ----------                  -----       ----  
** 
** is_tossing : 000000000000000000000000001111111111000000000000000000000000000000000000 
** was_tossing: 000000000000000000000000000000000001111111111111111111111111111111100000
** toss_timer :    =======================         ====================---- ... 321000==  
**                           
** Note 
** 0) When is_tossing is high a motion is reported/calculated.
** 1) if prevent_toss delay is 0 then toss to toss is allowed without pause.
** 2) if the pause is to short (< prevent_toss), the timers start over!
** 3) The 1 second limit is due to the touchpad only sends packets one second after a release.
** 
**
*/
static void tp_process_report (Gpm_Event *state,
				report_type report) 
{
  location_type loc;
  int           edges;
  float         pressure_speed_factor;
  float         edge_speed_factor;
  int           edge_motion_on;
  float         dx, dy;

  /* extract location and edges */
  loc.x = report.x;
  loc.y = report.y;
  edges = tp_edges (loc);


  if (report.fingers > 0) {

    if (tossing_enabled) {
      /* this is the cue to stop tossing, if we are tossing so
         Force a touch free pause before a new tossing is allowed (prevent toss time) */
      was_tossing = (was_tossing || is_tossing);
      toss_timer  = prevent_toss_packets;
      is_tossing  = 0;
      
      /* if we start tossing then this is from where */
      if (last_report.fingers == 0) {
	touch_loc = loc;
      }
    }

    /* Save the edge state, so we can detect a movement from non edge to edge. */
    if (!edges)
      was_non_edge = 1;

    /* Need enough packets to perform smoothing on dy and dx
    *  but it should work with only 2, so why is 4 required? */
    if (packet_num > 3) {

      /* Is this start of scrolling? */
      if (scrolling_enabled && !is_scrolling)
	is_scrolling  = ((edges & scrolling_edge)  && /* Note: only one &! */
			 !was_non_edge         && /* Must start at edge */
			 !last_report.gesture  && /* No consecutive tap */
			 (!corner_taps_enabled || /* Corner disabled or */
			  !is_corner (edges)));   /* no corner          */


      /* 
      ** 1) if edge motion is enabled, only activate if we moved into the edge or
      ** if not using the corners taps.  Dont activate right away after a gesture
      ** from a corner tap. 2) No harm is done if activated when scrolling (as
      ** long as the scrolling edge is vertical). 3)If user is tap holding (then
      ** he cannot lift the finger and hit the touchpad boarder).
      */

      edge_motion_on  = (edges && edge_motion_enabled &&
			  (was_non_edge         || 
			   !corner_taps_enabled ||
			  (!is_corner(edges) && !last_report.gesture)));
      edge_motion_on |= (edges && auto_scrolling_enabled && is_scrolling);
      edge_motion_on |= (edges && tap_hold_edge_motion_enabled && last_report.gesture && 
			 was_non_edge);


      /* compute the speed factor based on pressure */
      pressure_speed_factor = standard_speed_factor;

      if (report.pressure > speed_up_pressure) {
	pressure_speed_factor *= 1.0 + ((report.pressure - speed_up_pressure) *
			       speed_pressure_factor);
      }

      /* use the edge speed factor if edge_motion_speed is enabled */
      edge_speed_factor = (edge_motion_speed_enabled ?
			   pressure_speed_factor :
			   standard_speed_factor);
      
      pressure_speed_factor = (pressure_speed_enabled ?
			       pressure_speed_factor :
			       standard_speed_factor);
      
      if ( auto_scrolling_enabled && is_scrolling ){
	edge_speed_factor *= 1.0 + auto_scrolling_factor; 
      }


      /* Calculate dx and dy depending on whether we are at an edge */

      if (edge_motion_on && (edges & TOP_EDGE))
	 state->dy = -edge_speed * edge_speed_factor;
      else if (edge_motion_on && (edges & BOTTOM_EDGE))
	 state->dy = edge_speed * edge_speed_factor;
      else
	 state->dy = (pressure_speed_factor *
		      (((last_locs [mod4 (packet_num - 1)].y +
			 last_locs [mod4 (packet_num - 2)].y) / 2) - 
		       ((loc.y +
			 last_locs [mod4 (packet_num - 1)].y) / 2)));
      
      if (edge_motion_on && (edges & LEFT_EDGE))
	 state->dx = -edge_speed * edge_speed_factor;
      else if (edge_motion_on && (edges & RIGHT_EDGE))
	 state->dx = edge_speed * edge_speed_factor;
      else
	 state->dx = (pressure_speed_factor *
		      (((loc.x +
			 last_locs [mod4 (packet_num - 1)].x) / 2) - 
		       ((last_locs [mod4 (packet_num - 1)].x +
			 last_locs [mod4 (packet_num - 2)].x) / 2)));
    } /* If (packet_num > 3)*/

    last_locs [mod4 (packet_num)] = loc;
    was_edges = edges;
    packet_num++;
  } else {
    /* No finger on the pad */
    /* Start the tossing action if enabled */
    if (tossing_enabled && was_fingers == 1 &&
	!was_tossing    &&
	!is_scrolling   &&
	(packet_num > min_toss_packets) &&
	(packet_num < max_toss_packets)) {

      dx =    last_locs [mod4 (packet_num - 1)].x - touch_loc.x;
      dy = - (last_locs [mod4 (packet_num - 1)].y - touch_loc.y);

#if DEBUG_TOSS
      gpm_report (GPM_PR_INFO,"dx: %2.1f  dy: %2.1f  tdist^2: %2.1f",
		  dx, dy, distance (dx, dy));
#endif

      if (distance (dx, dy) > min_toss_dist__2) {
	is_tossing = 1;

	/* determine the toss speed */
	if (does_toss_use_static_speed) {
	  toss_speed.x = (static_toss_speed * standard_speed_factor *
			  dx / (abs (dx) + abs (dy)));
	  toss_speed.y = (static_toss_speed * standard_speed_factor *
			  dy / (abs (dx) + abs (dy)));
	} else {
	  toss_speed.x = dx * standard_speed_factor * toss_speed_factor;
	  toss_speed.y = dy * standard_speed_factor * toss_speed_factor;
	}

#if DEBUG_TOSS
	gpm_report (GPM_PR_INFO,"tossx: %d  tossy: %d", toss_speed.x, toss_speed.y);
#endif
      }
    }

    /* no fingers therefore: no edge, and restart packet count */
    was_non_edge = 0;
    is_scrolling = 0;
    scrolling_speed_timer = 0;
    packet_num = 0;
  }


  /* if we are tossing then apply the toss speed */
  if (tossing_enabled && is_tossing) {
    state->dx = toss_speed.x;
    state->dy = toss_speed.y;
    }


  /* if we are scrolling then stop moving and report wheel amount. */
  if ((scrolling_enabled && is_scrolling) ||
      is_always_scrolling){
    scrolling_amount_left += state->dy;
    state->dx = 0;
    state->dy = 0;
  }


  /* check for (corner)buttons if we didn't just complete a toss or is scrolling */
  if (!is_tossing && !was_tossing && !is_scrolling) {
    /*
    ** If there is no gesture then there are no buttons.
    ** If there is a continuing gesture then report the same buttons.
    ** Otherwise compute new buttons.
    */
    if (!report.gesture && 
	!report.left && !report.right && !report.middle && !report.fourth &&
	!report.up && !report.down ) {
      state->buttons = GPM_B_NONE;
    } else if (last_report.gesture) {
      state->buttons = last_state.buttons;
    } else {
      if (report.left)   state->buttons |= GPM_B_LEFT;
      
      if (report.right)  state->buttons |= GPM_B_RIGHT;

      if (report.middle) state->buttons |= GPM_B_MIDDLE;

      if (report.fourth) state->buttons |= GPM_B_FOURTH;

      if (report.up)  	 state->buttons |= GPM_B_UP;

      if (report.down)	 state->buttons |= GPM_B_DOWN;

      if ((tap_gesture_enabled && report.gesture && !is_corner (was_edges))){

	if( ( multi_finger_tap_enabled && (was_fingers == 0 || was_fingers == 1)) ||
	    (!multi_finger_tap_enabled && was_fingers > 0))
	  state->buttons |= GPM_B_LEFT;

	if(multi_finger_tap_enabled && was_fingers == 2)
	  state->buttons |= GPM_B_RIGHT;

	if(multi_finger_tap_enabled && was_fingers == 3)
	  state->buttons |= GPM_B_MIDDLE;

      }

      tp_process_corner_taps (state, report);
    }
  }

  /* Is there any amount of scrolling left?  Should be checked after corner actions. */
  if (scrolling_amount_left != 0){
    tp_handle_scrolling(state);
  }


  was_tossing  = was_tossing && toss_timer;
  if (was_tossing)
    toss_timer--;


  /* remember the last state of the finger for toss processing */
  last_report   = report;
  last_state    = *state;

  /* Dont do anything if the pad is not enabled, but after corner actions are
   * done so it can be turn on again. */
  if (!touchpad_enabled){
    state->buttons = GPM_B_NONE;
    state->dx      = 0;
    state->dy      = 0;
  }

#if DEBUG_REPORTS
  tp_dump_report_data (report, edges, state);
#endif

}


/*
** syn_read_config_file
**
** Read the configuration data from the global config file
** SYSCONFDIR "/gpm-syn.conf".
*/
void tp_read_config_file (char* config_filename) 
{
  char line [80];
  char *token;
  char *end_ptr;
  int param;
  FILE *config;
  char full_filename[100];
  int  status;

  status = snprintf(full_filename,100,SYSCONFDIR "/%s",config_filename);
  if (status < 0) {
    gpm_report (GPM_PR_WARN,"Too long path for configure file: %s", config_filename); 
    return;
  }


  if ( !(config = fopen (full_filename, "r")) ) {
    gpm_report (GPM_PR_WARN,"Failed to open configfile: %s", full_filename);    
  }else{
    while (fgets (line, 80, config)) {
      if (line [0] == '[') {
	if ( (token = strtok (line, "[] ")) ) {
	  param = 0;

	  /* which param is it */
	  while (param_data [param].name &&
		 strcasecmp (token, param_data [param].name) != 0) {
	    param++;
	  }

	  /* was a param found? */
	  if (!param_data [param].name) {
	    gpm_report (GPM_PR_WARN,"Unknown parameter %s", token);
	  } else {
	    token = strtok (NULL, "[] ");

	    switch (param_data [param].p_type) {
	    case Integer_Param:
	      *(param_data [param].addr.int_p) = strtol (token, &end_ptr, 0);
	      if (end_ptr == token)
		gpm_report (GPM_PR_WARN,"Value (%s) for parameter %s is invalid",
			    token, param_data [param].name);
#             if DEBUG_PARAMS
	      gpm_report (GPM_PR_INFO,"Param %s set to %d",
			  param_data [param].name,
			  *(param_data [param].addr.int_p));
#             endif
	      break;

	    case Float_Param:
	      *(param_data [param].addr.float_p) = strtod (token, &end_ptr);
	      if (end_ptr == token)
		gpm_report (GPM_PR_WARN,"Value (%s) for parameter %s is invalid",
			    token, param_data [param].name);
#             if DEBUG_PARAMS
	      gpm_report (GPM_PR_INFO,"Param %s set to %3.3f",
			  param_data [param].name,
			  *(param_data [param].addr.float_p));
#             endif
	      break;

	    case Flag_Param:
	      if (index ("YyTt1", token [0])) {
		*(param_data [param].addr.flag_p) = 1;
	      } else if (index ("NnFf0", token [0])) {
		*(param_data [param].addr.flag_p) = 0;
	      } else {
		gpm_report (GPM_PR_WARN,"Value (%s) for parameter %s is invalid",
			    token, param_data [param].name);
	      }
#             if DEBUG_PARAMS
	      gpm_report (GPM_PR_INFO,"Param %s set to %s",
			  param_data [param].name,
			  (*(param_data [param].addr.flag_p) ?
			   "True" : "False"));
#             endif
	      break;

	    case Action_Param:
	      param_data [param].addr.corner_p->action = strtol (token,
								 &end_ptr, 0);
	      if (end_ptr == token)
		gpm_report (GPM_PR_WARN,"Value (%s) for parameter %s is invalid",
			    token, param_data [param].name);
#             if DEBUG_PARAMS
	      gpm_report (GPM_PR_INFO,"Param %s set to %d",
			  param_data [param].name,
			  param_data [param].addr.corner_p->action);
#             endif
	      break;

	    default:
	    }
	  }
	}
      }
    }

    fclose (config);
  }
}


/*
** syn_process_touchpad_config
**
** Extract important information and report (as desired) to the user.
*/
static void syn_process_config (info_type ident,
				model_id_type model) 
{
  sensor = syn_get_sensor_info (model.info_sensor);
  gpm_report (GPM_PR_INFO, "     Firmware version %d.%d\n",
	      ident.info_major, ident.info_minor);

  tp_read_config_file ("gpm-syn.conf");

  /* Save important information */
  x_per_mm = sensor->x_per_mm;
  y_per_mm = sensor->y_per_mm;
  res_x = (int) (x_per_mm * sensor->width_mm);
  res_y = (int) (y_per_mm * sensor->height_mm);

  /* convert the tap times to packets (80 pkts/sec and 1000 ms/sec) */
  tap_lower_limit_packet = tap_lower_limit * 80 / 1000;
  tap_upper_limit_packet = tap_upper_limit * 80 / 1000;

  /* Convert the toss dist to touchpad resolution from mm */
  min_toss_dist__2 = sqr (min_toss_dist * (x_per_mm + y_per_mm) / 2);

  /* convert the toss times to packets (80 pkts/sec and 1000 ms/sec) */
  max_toss_packets     = max_toss_time * 80 / 1000;
  min_toss_packets     = min_toss_time * 80 / 1000;
  prevent_toss_packets = prevent_toss_time * 80 / 1000;

#if DEBUG
  gpm_report (GPM_PR_INFO,"Synaptics Ident:  model_code=%d   Firmware version %d.%d",
	      ident.info_model_code, ident.info_major, ident.info_minor);
  gpm_report (GPM_PR_INFO,"Synaptics model:");
  gpm_report (GPM_PR_INFO,"   rot180:      %s", model.info_rot180 ? "Yes" : "No");
  gpm_report (GPM_PR_INFO,"   portrait:    %s", model.info_portrait ? "Yes" : "No");
  gpm_report (GPM_PR_INFO,"   sensor:      %d", model.info_sensor);
  gpm_report (GPM_PR_INFO,"                %s", syn_model_name (model.info_sensor));
  gpm_report (GPM_PR_INFO,"                %s", sensor->model);
  gpm_report (GPM_PR_INFO,"                %dx%d res/mm",
	      sensor->x_per_mm, sensor->y_per_mm);
  gpm_report (GPM_PR_INFO,"                %4.1fx%4.1f mm",
	      sensor->width_mm, sensor->height_mm);
  gpm_report (GPM_PR_INFO,"                %dx%d res", res_x, res_y);
  gpm_report (GPM_PR_INFO,"   hardware:    %d", model.info_hardware);
  gpm_report (GPM_PR_INFO,"   newABS:      %s", model.info_new_abs ? "Yes" : "No");
  gpm_report (GPM_PR_INFO,"   simpleCmd:   %s", model.info_simple_cmd ? "Yes" : "No");
  gpm_report (GPM_PR_INFO,"   geometry:    %d", model.info_geometry);
  gpm_report (GPM_PR_INFO,"   extended:    %s", capabilities.cap_ext ? "Yes" : "No");
  gpm_report (GPM_PR_INFO,"   stick:       %s", capabilities.cap_stick ? "Yes" : "No");
  gpm_report (GPM_PR_INFO,"   sleep:       %s", capabilities.cap_sleep ? "Yes" : "No");
  gpm_report (GPM_PR_INFO,"   4 buttons:   %s", capabilities.cap_four_button ? "Yes" : "No");
  gpm_report (GPM_PR_INFO,"   multifinger: %s", capabilities.cap_multi_finger ? "Yes" : "No");
  gpm_report (GPM_PR_INFO,"   palmdetect:  %s", capabilities.cap_palm_detect ? "Yes" : "No");

#endif
#if DEBUG_TOSS
  gpm_report (GPM_PR_INFO,"\nMin Toss Dist^2: %d\n", min_toss_dist__2);
#endif
}


/****************************************************************************
**
** ROUTINES for interfacing to a SERIAL touchpad
**
****************************************************************************/




static unsigned char tp_hextoint (unsigned char byte1,
			       unsigned char byte2) 
{
  unsigned char bytes [3];
  int result;

  bytes [0] = byte1;
  bytes [1] = byte2;
  bytes [2] = '\0';
  sscanf (bytes, "%x", &result);
  return result;
}

static void tp_serial_flush_input (int fd) 
{
  struct timeval tv;
  fd_set rfds;
  unsigned char junk;
  
  FD_ZERO(&rfds);
  FD_SET (fd, &rfds);
  tv.tv_sec = 0;
  tv.tv_usec = 0;
  usleep (50000);

  while (select (fd+1, &rfds, NULL, NULL, &tv) == 1) {
#if DEBUG_FLUSH
    gpm_report (GPM_PR_INFO,"Serial tossing");
    fflush (stdout);
#endif
    read (fd, &junk, 1);
#if DEBUG_FLUSH
    gpm_report (GPM_PR_INFO," %c", junk);
#endif
    FD_ZERO(&rfds);
    FD_SET (fd, &rfds);
    tv.tv_sec = 0;
    tv.tv_usec = 0;
  }
}

static void tp_serial_read (int fd,
		      unsigned char *bytes,
		      int count) 
{
  struct timeval tv;
  fd_set rfds;
  int num_read = 0;
  int read_count;

  FD_ZERO (&rfds);
  FD_SET (fd, &rfds);
  tv.tv_sec = 1;
  tv.tv_usec = 0;

  while ((select (fd+1, &rfds, NULL, NULL, &tv) == 1) &&
	 (num_read < count)) {
    read_count = read (fd, &bytes [num_read], count - num_read);
    num_read += read_count;

    FD_ZERO (&rfds);
    FD_SET (fd, &rfds);
  }

  for (; num_read < count; num_read++) {
     bytes [num_read] = '\0';
  }
}

/* Write a string of commands */
static void tp_serial_send_cmd(int fd,
			 unsigned char *cmd) 
{
  unsigned char junk [15];

  tp_serial_flush_input (fd);
  write (fd, cmd, strlen (cmd));
  tp_serial_read (fd, junk, strlen (cmd));
#if DEBUG_FLUSH
  junk [strlen (cmd)] = '\0';
  gpm_report (GPM_PR_DEBUG,"serial tossing: %s", junk);
#endif
}

/* write 'mode' to a serial touchpad, STIG 58 */
static void syn_serial_set_mode (int fd,
				 unsigned char mode) 
{
  unsigned char bytes [15];

  sprintf (bytes, "%%C3B%02X5555", mode);
#if DEBUG_SENT_DATA
  gpm_report (GPM_PR_DEBUG,"modes: %s", bytes);
#endif
  tp_serial_send_cmd (fd, bytes);
}

/* read the identification from the serial touchpad, STIG 57*/
static void syn_serial_read_ident (int fd,
				   info_type *info) 
{
  unsigned char bytes [5];

  tp_serial_send_cmd (fd, "%A");
  tp_serial_read (fd, bytes, 4);

#if DEBUG_SENT_DATA
  bytes [4] = '\0';
  gpm_report (GPM_PR_INFO,"Serial ident: %s", bytes);
#endif

  if ( bytes [0] == 'S' && bytes [0] == 'T'){
    /* reformat the data
     * The out commented is wrong according to STIG page 57.*/
    /* info->info_model_code = (bytes [2] & 0x07) >> 3; */
    info->info_model_code = bytes [2] >> 3;
    info->info_major      = (bytes [2] & 0x07);
    info->info_minor      = bytes [3];
  }else{
    gpm_report (GPM_PR_ERR,"PS/2 serial device doesn't appear to be a synaptics touchpad\n");
    info->info_model_code = 0;
    info->info_minor      = 0;
    info->info_major      = 0;
  }
}



/* read the model_id from the serial touchpad (in ps/2 format) */
static void syn_serial_read_model_id (int fd,
				      model_id_type *model) 
{
  unsigned char bytes [7];
  int model_int;

  /* 
   * for older touchpads this command is not supported and no response will
   * come.  We should do non blocking input here to handle that case 
   * and return byte2 as 0x47 ... later.
   *
   * pebl: It is easier just to check version number less than 3.2, STIG page 60.
   */
  if ( (ident.info_major >= 4) ||
       (ident.info_major == 3 && ident.info_minor >= 2)){
    tp_serial_send_cmd (fd, "%D");
    tp_serial_read (fd, bytes, 6);

    /* reformat the data */
    model_int = ((tp_hextoint (bytes [0], bytes [1]) << 16) |
		 (tp_hextoint (bytes [2], bytes [3]) << 8) |
		 (tp_hextoint (bytes [4], bytes [5])));
    
#   if DEBUG_SENT_DATA
      bytes [6] = '\0';
      gpm_report (GPM_PR_DEBUG,"Serial model id: %s", bytes);
#   endif

  }else{
    model_int = 0;
  }

  syn_extract_model_id_info (model_int, model);
}


/* read the mode bytes and capabilities from the serial touchpad, STIG 57 */
static void syn_serial_read_cap (int fd,
				 ext_cap_type *cap) 
{
  unsigned char bytes [8];
  int cap_int = 0;

  tp_serial_send_cmd (fd, "%B");
  tp_serial_read (fd, bytes, 8);

#if DEBUG_SENT_DATA
  bytes [7] = '\0';
  gpm_report (GPM_PR_DEBUG,"Serial capabilites: %s", bytes);
#endif

  if (ident.info_major >= 4){
    if (bytes [0] == '3' && bytes [0] == 'B'){
      cap_int = ((tp_hextoint (bytes [4], bytes [5]) << 8) |
		 (tp_hextoint (bytes [6], bytes [7])));
    }else{
      gpm_report (GPM_PR_ERR,"PS/2 serial device doesn't appear to be a synaptics touchpad\n");
    }    
  }

  syn_extract_extended_capabilities(cap_int,cap);
}



/*------------------------------------------------------------------------*/
/*   PS/2 Utility functions.                                              */
/*      Adapted from tpconfig.c by C. Scott Ananian                       */
/*------------------------------------------------------------------------*/

/* PS2 Synaptics is using LSB, STIG page 29. 
**
** After power on or reset the touchpads are set to these defaults:
** 100 samples per second
** Resolution is 4 counts per mm
** Scaling 1:1
** Stream mode is selected
** Data reporting is disabled
** Absolute mode is disabled
*/


/* Normal ps2 commands, Command set is on STIG page 33  */
#define PS2_RESET         0xFF      /* Reset */
#define PS2_RESEND        0xFE      /* Resend command */
#define PS2_SET_DEFAULT   0xF6      /* Set default */
#define PS2_DISABLE_DATA  0xF5      /* Stop sending motion data packets */
#define PS2_ENABLE_DATA   0xF4      /* Start sending motion data packets */
#define PS2_SAMPLE_RATE   0xF3      /* Set sample rate to value in a following byte */
#define PS2_READ_DEVICE   0xF2      /* Read device type (ack and idcode) */
#define PS2_REMOTE_MODE   0xF0      /* Set touchpad in remote mode */
#define PS2_WRAP_MODE     0xEE      /* Enter wrap mode */
#define PS2_END_WRAP_MODE 0xEE      /* Leave wrap mode */
#define PS2_READ_DATA     0xEB      /* Read move data (in remote mode)*/
#define PS2_STREAM_MODE   0xEA      /* Enter stream mode (device sends move data) */
#define PS2_STATUS_REQ    0xE9      /* Ask for status request */
#define PS2_RESOLUTION    0xE8      /* Set resolution to next transmitted byte */
#define PS2_SCALE_12      0xE7      /* Set scale to 1:2 */
#define PS2_SCALE_11      0xE6      /* Set scale to 1:1 */


/* Normal ps2 responce */
#define PS2_ERROR         0xFC      /* Error, after a reset,resend or disconnect*/
#define PS2_ACK           0xFA      /* Command acknowledge */
#define PS2_READY         0xAA      /* Send after a calibration or ERROR */
#define PS2_MOUSE_IDCODE  0x00      /* Identification code (meaning mouse) sent after a PS2_READY */


/* Additional synaptic commands*/
#define PS2_SYN_CMD       0xE8      /* Four of these each with an following byte encodes a command*/
#define PS2_SYN_INERT     0xE6      /* This ps2 command is ignored by synaptics */
#define PS2_SYN_SET_MODE1 0x0A      /* Set the mode byte 1 instead of sample rate (used after a sample rate cmd) */
#define PS2_SYN_SET_MODE2 0x14      /* Set the mode byte 2 instead of sample rate (used after a sample rate cmd).
				     * All other sample rate gives undefined behavior (used to address 4 byte mode)*/
#define PS2_SYN_STATUS_OK 0x47      /* Special synaptics Status report is recognized */


/* These are the commands that can be given (encoded) by PS_SYN_CMD */
#define PS2_SYN_CMD_IDENTIFY     0x00 /* Identify Touchpad */
#define PS2_SYN_CMD_MODES        0x01 /* Read Touchpad Modes */
#define PS2_SYN_CMD_CAPABILITIES 0x02 /* Read capabilities */
#define PS2_SYN_CMD_MODEL_ID     0x03 /* Read model id */
#define PS2_SYN_CMD_SERIAL_NO_P  0x06 /* Read serial number prefix */
#define PS2_SYN_CMD_SERIAL_NO_S  0x07 /* Read serial number suffix */
#define PS2_SYN_CMD_RESOLUTIONS  0x08 /* Read resolutions */




typedef unsigned char byte;

/* read a byte from the ps/2 port */
static byte tp_ps2_getbyte(int fd) 
{
  byte b;

  read(fd, &b, 1);
  return b;
}


/* write a byte to the ps/2 port, handling ACK and resend.*/
static void tp_ps2_putbyte(int fd,
			byte b) 
{
  byte ack;

  do{
    write(fd, &b, 1);
    read(fd, &ack, 1);
  } while(ack == PS2_RESEND);
  
  if (ack != PS2_ACK)
    gpm_report (GPM_PR_ERR,"Invalid ACK to synps2 send byte %02X, got %02X",b,ack);
}

/* use the Synaptics extended ps/2 syntax to write a special command byte 
* STIG page 36: Send exactly four PS2_SYN_CMD (which is otherwise ignored)
* and after each a byte with the 2 first bits being the command (LSB). End it with
* either  PS2_SAMPLE_RATE or PS2_STATUS_REQ. It is hinted to send an inert command
* first so not having five or more PS2_SYN_CMD by coincident.
*/
static void syn_ps2_send_cmd(int fd,
			 byte cmd) 
{
  int i;

  /* initialize with 'inert' command */
  tp_ps2_putbyte(fd, PS2_SYN_INERT);
  for (i=0; i<4; i++) {
    tp_ps2_putbyte(fd, PS2_SYN_CMD);
    tp_ps2_putbyte(fd, (cmd>>6)&0x3);
    cmd<<=2;
  }
}

#if 0

/* write 'cmd' to mode byte 1.
 * This function is not used. 
 * Code 0x0A is unknown to me, maybe used in older synaptics?
 *
 * This is used for some old 2 byte control synaptics. The function is not
 * used, and is probably leftover from mixing with Van der Plas code.
 */
static void syn_ps2_set_mode1(int fd,
			  byte cmd) 
{
  syn_ps2_send_cmd(fd, cmd);
  tp_ps2_putbyte(fd, PS2_SAMPLE_RATE);
  tp_ps2_putbyte(fd, PS2_SYN_SET_MODE1);
}

#endif


/* write 'cmd' to mode byte 2 
 * See ps2_send_cmd. PS2_SR_SET_MODE stores the touchpad mode encoded in the 
 * four PS2_SYN_CMD commands
 */
static void syn_ps2_set_mode2(int fd,
			  byte cmd) 
{
  syn_ps2_send_cmd(fd, cmd);
  tp_ps2_putbyte(fd, PS2_SAMPLE_RATE);
  tp_ps2_putbyte(fd, PS2_SYN_SET_MODE2);
}



/* read three byte status ('a','b','c') corresponding to register 'cmd' 
*  Special status request for synaptics is given after a cmd.
*  Byte b is PS2_SYN_STATUS_OK to recognize a synaptics
*/
static void syn_ps2_status_rqst(int fd,
			    byte cmd,
			    byte *bytes) 
{
  syn_ps2_send_cmd(fd, cmd);
  tp_ps2_putbyte(fd, PS2_STATUS_REQ);
  bytes [0]=tp_ps2_getbyte(fd);
  bytes [1]=tp_ps2_getbyte(fd);
  bytes [2]=tp_ps2_getbyte(fd);
}


#if 0

/* read the modes from the touchpad (in ps/2 format) */
static void syn_ps2_read_modes (int fd) 
{
  unsigned char bytes [3];

  syn_ps2_status_rqst (fd, PS2_SYN_CMD_MODES, bytes);
# ifdef DEBUG
  gpm_report (GPM_PR_INFO,"Synaptic PS/2 touchpad modes: %02X", bytes [2]);
# endif
}

#endif


/* read the identification from the ps2 touchpad */
static void syn_ps2_read_ident (int fd,
				info_type *info) 
{
  byte bytes [3];

  syn_ps2_status_rqst (fd, PS2_SYN_CMD_IDENTIFY, bytes);
  if (bytes [1] != PS2_SYN_STATUS_OK) {
    gpm_report (GPM_PR_ERR,"PS/2 device doesn't appear to be a synaptics touchpad\n");
    info->info_minor      = 0;
    info->info_model_code = 0;
    info->info_major      = 0;

  } else {
    info->info_minor      = bytes [0];
    info->info_model_code = (bytes [2] >> 4) & 0x0f;
    info->info_major      = bytes [2] & 0x0f;
  }
}


/* read the model_id from the ps2 touchpad */
static void syn_ps2_read_model_id (int fd,
				   model_id_type *model) 
{
  unsigned char bytes [3];
  int model_int;

  syn_ps2_status_rqst (fd, PS2_SYN_CMD_MODEL_ID, bytes);
  model_int = ((bytes [0] << 16) |
	       (bytes [1] << 8)  |
	       (bytes [2]));
  syn_extract_model_id_info (model_int, model);
}




/* read the extended capability from the ps2 touchpad, STIG page 15 */
static void syn_ps2_read_cap (int fd,
                              ext_cap_type *cap)
{
  unsigned char bytes [3];
  int ext_cap_int;

  syn_ps2_status_rqst (fd, PS2_SYN_CMD_CAPABILITIES, bytes);

  if (bytes [1] != PS2_SYN_STATUS_OK) {
    gpm_report (GPM_PR_ERR,"PS/2 device doesn't appear to be a synaptics touchpad\n");
    ext_cap_int = 0;
  }else{
    ext_cap_int = bytes[0] << 8 | bytes[2];
  }

  syn_extract_extended_capabilities(ext_cap_int, cap);
}



/*
 * ps2_disable_data
 *
 * Disable data reporting (streaming), and flush eventual old packets. As the
 * kernel keeps a queue of received data from the touchpad, the next byte we
 * read could be old data and not the ack to our command or request.  Disable
 * data should always be called before sending commands or request to the
 * touchpad.
 *
 * Note that this is a general ps2 command which should not be in this file,
 * but in mice.c.
 */
static void tp_ps2_disable_data (int fd) 
{
  struct timeval tv;
  fd_set rfds;
  unsigned char status;
  byte cmd = PS2_DISABLE_DATA;

  
  FD_ZERO(&rfds);
  FD_SET (fd, &rfds);
  tv.tv_sec = 0;
  tv.tv_usec = 0;

  write(fd,&cmd,1);

  usleep (3000);
  
  while (select (fd+1, &rfds, NULL, NULL, &tv) == 1) {
    read (fd, &status, 1);
#if DEBUG_RESET
    gpm_report (GPM_PR_INFO,"PS/2 touchpad disable data flush: %02X", status);
#endif
    FD_ZERO(&rfds);
    FD_SET (fd, &rfds);
    tv.tv_sec = 0;
    tv.tv_usec = 0;
  }

  if (status != PS2_ACK)
    gpm_report (GPM_PR_ERR,"Invalid ACK to disable stream command, got %02X",status);
}


/* 
 * ps2_enable_data
 *
 * Enable data after a disable data command. Should have called the disable data
 * before calling this function.
 */

static void tp_ps2_enable_data(int fd)
{  
  tp_ps2_putbyte(fd,PS2_ENABLE_DATA);

}


/* 
 * ps2_send_reset
 *
 * Send reset command and absorb additional READY, IDCODE from the
 * touchpad. Should have called the disable data before calling this function.
 * Synaptics garanties always to return PS2_READY. STIG page 31 and 48.
 */

static void tp_ps2_send_reset(int fd)
{
  byte status,id_code = PS2_MOUSE_IDCODE;
  byte reset_cmd = PS2_RESET;
  
  /* Send reset command. */
  write(fd,&reset_cmd,1);
  status = tp_ps2_getbyte(fd);

  /* Sometimes the touchpad sends additional ready,idcode before ack the reset command.
   * I dont know why! */
  while (status == PS2_READY){
    gpm_report(GPM_PR_INFO,"PS/2 Touchpad sending additional READY, ID CODE.\n");
    id_code = tp_ps2_getbyte(fd);
    status  = tp_ps2_getbyte(fd);
  }

  if (status != PS2_ACK || id_code != PS2_MOUSE_IDCODE){
    gpm_report(GPM_PR_ERR,"Sending reset command to PS/2 TouchPad failed: No ACK, got %02X.\n",status);
  }
    
  /* Wait 750 ms to recalibrate. */
  usleep(7500);

  if ( (status  = tp_ps2_getbyte(fd)) != PS2_READY ){
    gpm_report(GPM_PR_ERR,"Reseting PS/2 TouchPad failed: No READY, got %02X.\n",status);
  }
  if ( (id_code = tp_ps2_getbyte(fd)) != PS2_MOUSE_IDCODE){
    gpm_report(GPM_PR_ERR,"Reseting PS/2 TouchPad failed: Wrong ID, got %02X.\n",id_code);
  }

}



/*
** syn_ps2_absolute_mode
**
** Put the touchpad into absolute mode.
*/

static void syn_ps2_absolute_mode(int fd)
{
  /* select 6 byte packet, high packet rate, no-sleep */
  syn_ps2_set_mode2 (fd, (ABSOLUTE_MODE    |
		      HIGH_REPORT_RATE |
		      PS2_NO_SLEEP     |
		      (stick_enabled ? STICK_ENABLED : STICK_DISABLE) |
		      (wmode_enabled ? REPORT_W_ON : REPORT_W_OFF)));
  
}



/****************************************************************************
**
**  TRANSLATE FUNCTIONS the incoming data to an uniform report
**
****************************************************************************/

/* SERIAL PACKETS
 * STIG page 63, note 7 bit!
 *
 * byte 0 |    1   | reserved | gesture  | finger |  left   |  middle | right|  
 * byte 1 |    0   |                     x-pos 7-12                          |
 * byte 2 |    0   |                     x-pos 1-6                           |
 * byte 3 |    0   |                     y-pos 7-12                          |
 * byte 4 |    0   |                     y-pos 1-6                           |
 * byte 5 |    0   |                  z pressure 7-2                         |
 * byte 6 |    0   |  down    |   up     |y-pos 0 | x-pos 0 | z pressure 0-1 |  (new_abs set)
 * byte 7 |    0   |      reserved       |       W mode 0-3                  |  (new_abs and wmode set)
 */


static void syn_serial_translate_data (unsigned char *data,
					 report_type *report) 
{
  report->gesture  = check_bits (data [0], 0x10);
  report->fingers  = check_bits (data [0], 0x08);
  report->left     = check_bits (data [0], 0x04);
  report->middle   = check_bits (data [0], 0x02);
  report->right    = check_bits (data [0], 0x01);
  report->x        = (data [1] << 7) | (data [2] << 1);
  report->y        = (data [3] << 7) | (data [4] << 1);
  report->pressure = data [5] << 2;
  report->fourth   = 0;
  report->up       = 0;
  report->down     = 0;
  report->w        = 0;

  if (model.info_new_abs){
    report->up        = check_bits (data [6], 0x20);
    report->down      = check_bits (data [6], 0x40);
    report->y        |= (data [6] & 0x08) >> 3;
    report->x        |= (data [6] & 0x04) >> 2;
    report->pressure |= (data [6] & 0x03);

    if (wmode_enabled){
      report->w         = (data [7] & 0x0F);
    }
  }
}


/* PS2 PACKETS 
 *
 * Handle error packets. It may be garbage or not, as the synaptics pad keeps sending data 1
 * sec after last touch.
 */

static void syn_ps2_translate_error(unsigned char *data,
				    report_type *report)
{
  gpm_report(GPM_PR_WARN,"Unrecognized Synaptic PS/2 Touchpad packet: %02X %02X %02X %02X %02X %02X",
	      data [0],data [1],data [2],data [3],data [4],data [5]);

  if (reset_on_error_enabled) {
    /* Hack to get the fd: which_mouse is the current mouse,
       and as the synaptic code is called, it is the current mouse. */
    syn_ps2_reset(which_mouse->fd);
    syn_ps2_absolute_mode(which_mouse->fd);
  }
  
  report->left     = 0;
  report->middle   = 0;
  report->right    = 0;
  report->fourth   = 0;
  report->up       = 0;
  report->down     = 0;
  report->x        = 0;
  report->y        = 0;
  report->pressure = 0;
  report->gesture  = 0;
  report->fingers  = 0;
  report->w        = 0;
  
}


/*
 * STIG page 42
 * wmode = 0, newer version. Gesture, right and left are repeated.
 *
 * byte 0 |     1   |    0   |  Finger  | Reserved |    0   | Gesture |  Right  | Left |  
 * byte 1 |         y-pos 11-8                     |         x-pos  11-8               |
 * byte 2 |                           z pressure 0-7                                   |
 * byte 3 |     1   |    1   | y-pos 12 | x-pos 12 |    0   | Gesture |  Right  | Left |
 * byte 4 |                                 x - pos 0-7                                |
 * byte 5 |                                 y - pos 0-7                                |
 *
 * STIG page 43
 * wmode = 0, old version <  3.2.
 * Second is a second gesture!?
 *
 * byte 0 |     1   |    1   |  z-pres 6-7         | Second | Gesture |  Right  | Left |  
 * byte 1 |  finger |    0   |     0    |                    x-pos  12-8               |
 * byte 2 |                           x-pos  0-7                                       |
 * byte 3 |     1   |    0   |               z-pressure 0-5                            |
 * byte 4 |Reserved |    0   |    0     |              y - pos 8-12                    |
 * byte 5 |                                 y - pos 0-7                                |
 *
 
 */

/* Translate the reported data into a record for processing */
static void syn_ps2_translate_data (unsigned char *data,
				    report_type *report) 
{
  
  /* Check that this is indeed an absolute 6 byte new version packet*/
  if (((data [0] & 0xc8) == 0x80) &&              /* Check static in byte 0 */
      ((data [3] & 0xc8) == 0xc0) &&              /* Check static in byte 3 */
      ((data [0] & 0x0F) == (data [3] & 0x0F))) { /* check repeated data */
    report->left     = check_bits (data [0], 0x01);
    report->middle   = 0;
    report->right    = check_bits (data [0], 0x02);
    report->fourth   = 0;
    report->up       = 0;
    report->down     = 0;
    report->x        = (((data [1] & 0x0F) << 8) |
			((data [3] & 0x10) << 8) |
			((data [4])));
    report->y        = (((data [1] & 0xF0) << 4) |
			((data [3] & 0x20) << 7) |
			((data [5])));
    report->pressure = data [2];
    report->gesture  = check_bits (data [0], 0x04);
    report->fingers  = check_bits (data [0], 0x20);
    report->w        = 0;
    
  } /* Old style packet maybe */
  else if (((data [0] & 0xC0) == 0xC0) && /* Static in byte 0*/
	   ((data [1] & 0x60) == 0x00) && /* Static in byte 1*/
	   ((data [3] & 0xC0) == 0x80) && /* Static in byte 3*/
	   ((data [4] & 0x60) == 0x00)) { /* Static in byte 4*/
    report->left     = check_bits (data [0], 0x01);
    report->middle   = 0;
    report->right    = check_bits (data [0], 0x02);
    report->fourth   = 0;
    report->up       = 0;
    report->down     = 0;
    report->x        = (((data [1] & 0x1F) << 8) |
			((data [2])));
    report->y        = (((data [4] & 0x1f) << 8) |
			((data [5])));
    report->pressure = (((data [0] & 0x30) << 2 ) |
			((data [3] & 0x3f)));
    report->gesture  = check_bits (data [0], 0x04);
    report->fingers  = check_bits (data [1], 0x80);
    report->w        = 0;
    
  } else { 
    syn_ps2_translate_error(data,report);
  }
}


/* STIG page 42
 * wmode = 1, 
 *
 * byte 0 |     1   |    0   |  W 2-3              |    0   | W 1     |  Right  | Left |  
 * byte 1 |         y-pos 11-8                     |         x-pos  11-8               |
 * byte 2 |                           z pressure 0-7                                   |
 * byte 3 |     1   |    1   | y-pos 12 | x-pos 12 |    0   | W 0     |  R/D    | L/U  |
 * byte 4 |                                 x - pos 0-7                                |
 * byte 5 |                                 y - pos 0-7                                |
 *
 */

static void syn_ps2_translate_wmode_data (unsigned char *data,
					    report_type *report) 
{
  /* Check that it is an absolute packet */
  if (((data[0] & 0xc8) == 0x80) && ((data[3] & 0xc8) == 0xc0)) {

    report->left     = check_bits (data[0], 0x01);
    report->middle   = check_bits (data[0] ^ data[3], 0x01);
    report->right    = check_bits (data[0], 0x02);
    report->fourth   = check_bits (data[0] ^ data[3], 0x02);
    report->up       = 0;
    report->down     = 0;
    report->x        = (((data[1] & 0x0F) << 8) |
			((data[3] & 0x10) << 8) |
			((data[4])));
    report->y        = (((data[1] & 0xF0) << 4) |
			((data[3] & 0x20) << 7) |
			((data[5])));
    report->pressure = data[2];
    report->fingers  = 0;
    report->gesture  = 0;
    report->w = (((data[3] & 0x04) >> 2) |
		 ((data[0] & 0x04) >> 1) |
		 ((data[0] & 0x30) >> 2));


  } else { 
    syn_ps2_translate_error(data,report);
  }
}


/****************************************************************************
**
** INTERFACE ROUTINES
**
****************************************************************************/

/*
** syn_process_serial_data
**
** Process the touchpad 6 byte report.
*/
void syn_process_serial_data (Gpm_Event *state,
			      unsigned char *data) 
{
  /* initialize the state */
  state->buttons = 0;
  state->dx      = 0;
  state->dy      = 0;

  syn_serial_translate_data (data, &report);
  if (tp_preprocess_report(&report,state)) return;
  if (wmode_enabled){
    tp_process_wmode_report(&report);    
  }    

  tp_process_report (state, report);
}


/*
** syn_serial_reset
**
** Reset the touchpad to relative mode.  This cannot be called directly as a
** command should be sent in 1200 baud, not 9600.
*/

void syn_serial_reset(int fd)
{
  gpm_report (GPM_PR_INFO,"Reseting Synaptic Serial Touchpad.");

  syn_serial_set_mode (fd, (RELATIVE_MODE    |
			    HIGH_REPORT_RATE |
			    USE_9600_BAUD    |
			    NORMAL_REPORT    |
			    REPORT_W_OFF));

}

/*
** syn_serial_init
** 
** Initialize the synaptics touchpad.  Read model and identification.
** Determine the size of the touchpad in "pixels".  Select 6/7/8 byte packets,
** select 9600 baud, and select high packet rate.
*/
int syn_serial_init (int fd) 
{
  int return_packetlength;

  gpm_report(GPM_PR_DEBUG,"Initializing Synaptics Serial TouchPad");

  syn_serial_read_ident (fd, &ident);
  syn_serial_read_model_id (fd, &model);
  syn_serial_read_cap(fd, &capabilities);

  syn_process_config (ident, model);

  /* Change the protocol to use either 6,7 or 8 bytes, STIG 63 */
  if (model.info_new_abs){
    if (wmode_enabled){
      return_packetlength = 8;
    }else{
      return_packetlength = 7;
    }
  }else{
    return_packetlength = 6;
    wmode_enabled = 0;
  }

  syn_serial_set_mode (fd, (ABSOLUTE_MODE    |
			    HIGH_REPORT_RATE |
			    USE_9600_BAUD    |
			    (model.info_new_abs ? EXTENDED_REPORT : NORMAL_REPORT) |
			    (wmode_enabled ? REPORT_W_ON : REPORT_W_OFF)));

  return return_packetlength;
}


/*
** syn_process_ps2_data
**
** Process the touchpad 6 byte report.
*/
void syn_process_ps2_data (Gpm_Event *state,
			   unsigned char *data) 
{
  /*   gpm_report(GPM_PR_DEBUG,"Data %02x %02x %02x %02x %02x %02x",data[0],data[1],data[2],data[3],data[4],data[5]); */

  /* initialize the state */
  state->buttons = 0;
  state->dx      = 0;
  state->dy      = 0;


  if (wmode_enabled) {
    syn_ps2_translate_wmode_data (data, &report);
    if (syn_ps2_process_extended_packets(data,&report,state)) return;      
    if (tp_preprocess_report(&report,state)) return;
    tp_process_wmode_report(&report);
  }else {
    syn_ps2_translate_data (data, &report);
    if (tp_preprocess_report(&report,state)) return;
  }

  tp_process_report (state, report);
}




/*
** syn_serial_reset
**
** Reset the touchpad and set to relative mode (ps/2).
*/

void syn_ps2_reset (int fd)
{
  gpm_report (GPM_PR_INFO,"Reseting Synaptic PS/2 Touchpad.");

  /* Stop incoming motion data (of whatever kind absolute/relative). */
  tp_ps2_disable_data(fd);
  tp_ps2_send_reset(fd);
  tp_ps2_enable_data(fd);

}


/*
** syn_ps2_init
** 
** Initialize the synaptics touchpad.  Read model and identification.
** Determine the size of the touchpad in "pixels".  Select 6 byte packets,
** and select high packet rate.
*/
void syn_ps2_init (int fd) 
{

  gpm_report(GPM_PR_DEBUG,"Initializing Synaptics PS/2 TouchPad");

  tp_ps2_disable_data(fd);

  tp_ps2_send_reset(fd);

  syn_ps2_read_ident    (fd, &ident);
  syn_ps2_read_model_id (fd, &model);
  syn_ps2_read_cap      (fd, &capabilities);

  syn_process_config (ident, model);
  syn_ps2_absolute_mode(fd);

  tp_ps2_enable_data(fd);
}
